import json
import os

from .abstract_dataset import AbstractDataset
from ForensicHub.registry import register_dataset


@register_dataset('ManiDataset')
class ManiDataset(AbstractDataset):
    """Dataset for loading images from a directory structure with Tp and Gt folders.
    
    The directory structure should be:
    path/
        Tp/
            image1.jpg
            image2.jpg
            ...
        Gt/
            mask1.jpg
            mask2.jpg
            ...
    """

    def _init_dataset_path(self):
        # 检查目录结构
        if not os.path.exists(self.path):
            raise FileNotFoundError(f"dataset path {self.path} not found")

        tp_dir = os.path.join(self.path, "Tp")
        gt_dir = os.path.join(self.path, "Gt")

        if not os.path.exists(tp_dir):
            raise FileNotFoundError(f"test images directory {tp_dir} not found")
        if not os.path.exists(gt_dir):
            raise FileNotFoundError(f"ground truth directory {gt_dir} not found")

        # 获取所有图片路径
        self.tp_path = []
        self.gt_path = []

        for filename in os.listdir(tp_dir):
            if filename.endswith(('.jpg', '.jpeg', '.png')):
                tp_path = os.path.join(tp_dir, filename)
                gt_path = os.path.join(gt_dir, filename)

                if not os.path.exists(gt_path):
                    raise FileNotFoundError(f"ground truth {gt_path} not found")

                self.tp_path.append(tp_path)
                self.gt_path.append(gt_path)

        return self.tp_path, self.gt_path


@register_dataset('JsonDataset')
class JsonDataset(AbstractDataset):
    """Dataset for loading images from a JSON file.
    
    The JSON file should be organized as:
    [
        ["./Tp/6.jpg", "./Gt/6.jpg"],
        ["./Tp/7.jpg", "./Gt/7.jpg"],
        ["./Tp/8.jpg", "Negative"],
        ...
    ]
    If the second element is "Negative", it means the image is a negative sample
    with no manipulation.
    """

    def _init_dataset_path(self):
        # 读取json文件
        with open(self.path, 'r') as f:
            data = json.load(f)

        # 检查json格式
        if not isinstance(data, list):
            raise ValueError("json file should be a list of pairs")

        # 初始化路径列表
        self.tp_path = []
        self.gt_path = []

        # 遍历json数据
        for pair in data:
            if not isinstance(pair, list) or len(pair) != 2:
                raise ValueError("each element in json file should be a pair")

            tp_path = pair[0]
            gt_path = pair[1]

            # 检查文件是否存在
            if not os.path.exists(tp_path):
                raise FileNotFoundError(f"test image {tp_path} not found")

            if gt_path != "Negative" and not os.path.exists(gt_path):
                raise FileNotFoundError(f"ground truth {gt_path} not found")

            self.tp_path.append(tp_path)
            self.gt_path.append(gt_path)

        return self.tp_path, self.gt_path

    def __str__(self) -> str:
        """Return a string representation of the dataset."""
        return f"{self.__class__.__name__}({len(self)} samples)"



import json
import os
import random
import torch
import numpy as np
from PIL import Image
from ForensicHub.core.base_dataset import BaseDataset
from ForensicHub.registry import register_dataset


@register_dataset("DocumentCrossDataset")
class DocumentCrossDataset(BaseDataset):
    def __init__(self, path, image_size=224, **kwargs):
        self.image_size = image_size
        super().__init__(path=path, **kwargs)

    def _init_dataset_path(self) -> None:
        """Read dataset directories and parse image paths, masks, and labels."""
        all_samples = []

        path_list = [self.path] if isinstance(self.path, str) else self.path

        for path_id, one_path in enumerate(path_list):
            if not os.path.exists(one_path):
                raise FileNotFoundError(f"Dataset folder not found at {one_path}")

            image_dir = os.path.join(one_path, 'images')
            mask_dir = os.path.join(one_path, 'masks')

            if not os.path.exists(image_dir) or not os.path.exists(mask_dir):
                raise FileNotFoundError(f"Missing 'images' or 'masks' folder in {one_path}")

            image_list = sorted(os.listdir(image_dir))
            mask_list = sorted(os.listdir(mask_dir))

            assert len(image_list) == len(mask_list), \
                f"Mismatch in number of images and masks in {one_path}"

            for i in range(len(image_list)):
                sample = {
                    'path': os.path.join(image_dir, image_list[i]),
                    'mask': os.path.join(mask_dir, mask_list[i]),
                    'label': None,
                    # 'index': i,
                    # 'source': one_path  # 可选：记录来源路径
                }
                all_samples.append(sample)

        self.samples = all_samples
        self.entry_path = self.path  # For __str__()

    def __len__(self):
        return len(self.samples)

    # def _get_random_samples(self) -> list:
    #     """每个epoch随机选择pic_num张数据。"""
    #     if self.pic_num >=0:
    #         return random.sample(self.samples, self.pic_num)
    #     else:
    #         self.samples

    def __getitem__(self, idx):
        sample = self.samples[idx]
        image_path = sample["path"]
        mask_path = sample["mask"]
        label = sample["label"]

        image = Image.open(image_path).convert("RGB")
        image = image.resize((self.image_size, self.image_size))
        image = np.array(image)

        mask = Image.open(mask_path).convert("L")  # 'L'模式表示灰度图
        mask = mask.resize((self.image_size, self.image_size), resample=Image.NEAREST)
        mask = np.array(mask)
        mask = mask / 255.

        # Apply transforms
        if self.common_transform:
            out = self.common_transform(image=image, mask=mask)
            image = out['image']
            mask = out['mask']
        if self.post_transform:
            out = self.post_transform(image=image, mask=mask)
            image = out['image']
            mask = out['mask']
        mask = mask.float()
        output = {
            "image": image,
            "mask": mask,
            "label": torch.tensor(1).long() if torch.sum(mask) > 0 else torch.tensor(0).long()
        }

        # Apply post-processing functions
        if self.post_funcs:
            self.post_funcs(output)

        return output

    def __str__(self):
        return (f"DocumentCrossDataset from: {self.entry_path}\n"
                f"Total samples: {len(self.samples)}\n")


if __name__ == '__main__':
    # 这是pixel level 的mask的训练集参数 全假图  测试集对应路径旁边的test

    dataset = DocumentCrossDataset(path=['/mnt/data1/public_datasets/Doc/DocTamperV1-TrainingSet_reset_to_realworld',
                                         '/mnt/data1/public_datasets/Doc/cutted_datasets_fakes/OSTF_train',
                                         '/mnt/data1/public_datasets/Doc/cutted_datasets_fakes/RealTextManipulation_train',
                                         '/mnt/data1/public_datasets/Doc/cutted_datasets_fakes/T-SROIE_train',
                                         '/mnt/data1/public_datasets/Doc/cutted_datasets_fakes/Tampered-IC13_train'])
    # 总共七个测试集 DocTamperV有三个测试集 剩下四个数据集 分别是有一个
    # Doctamper的三个测试集分别如下 /mnt/data1/public_datasets/Doc/DocTamperV1/DocTamperV1-TestingSet /mnt/data1/public_datasets/Doc/DocTamperV1/DocTamperV1-FCD /mnt/data1/public_datasets/Doc/DocTamperV1/DocTamperV1-SCD
    # 剩下的为
    # /mnt/data1/public_datasets/Doc/cutted_datasets_fakes/OSTF_test
    # /mnt/data1/public_datasets/Doc/cutted_datasets_fakes/RealTextManipulation_test
    # /mnt/data1/public_datasets/Doc/cutted_datasets_fakes/T-SROIE_test
    # /mnt/data1/public_datasets/Doc/cutted_datasets_fakes/Tampered-IC13_test

    # 如果做image-level
    # 那就是把cutted_datasets_fakes换成cutted_datasets_alls  也就是全集 但是Doctamper仍然是只有假图 也训练带上
    # path=['/mnt/data1/public_datasets/Doc/DocTamperV1-TrainingSet_reset_to_realworld','/mnt/data1/public_datasets/Doc/cutted_datasets_alls/OSTF_train','/mnt/data1/public_datasets/Doc/cutted_datasets_alls/RealTextManipulation_train','/mnt/data1/public_datasets/Doc/cutted_datasets_alls/T-SROIE_train','/mnt/data1/public_datasets/Doc/cutted_datasets_alls/Tampered-IC13_train']
    # 测试集如下
    # Doctamper的三个测试集分别如下 /mnt/data1/public_datasets/Doc/DocTamperV1/DocTamperV1-TestingSet /mnt/data1/public_datasets/Doc/DocTamperV1/DocTamperV1-FCD /mnt/data1/public_datasets/Doc/DocTamperV1/DocTamperV1-SCD
    # 剩下的为
    # /mnt/data1/public_datasets/Doc/cutted_datasets_alls/OSTF_test
    # /mnt/data1/public_datasets/Doc/cutted_datasets_alls/RealTextManipulation_test
    # /mnt/data1/public_datasets/Doc/cutted_datasets_alls/T-SROIE_test
    # /mnt/data1/public_datasets/Doc/cutted_datasets_alls/Tampered-IC13_test
    dataset.__getitem__(0)
    import pdb;

    pdb.set_trace()



from .noise.noise_3down import *
from .rgb.rgb_3down import *
from .crosstrans2 import *
import torch.nn as nn

from ForensicHub.core.base_model import BaseModel
from ForensicHub.registry import register_model

'''
Ai-generated image detection using a cross-attention enhanced dual-stream network
'''


@register_model("DualNet")
class DualNet(BaseModel):
    def __init__(self):
        super(DualNet, self).__init__()
        self.noise = Trans_Noise()
        self.rgb = Pre2()
        self.ct = crosstrans(depth=2, dim=256, hidden_dim=1024, heads=4, head_dim=64, dropout=0.1)
        self.n_elayers = nn.Sequential(
            residual(256, 256),
            residual(256, 256)
        )
        self.r_elayers = nn.Sequential(
            plain(256, 256),
            plain(256, 256)
        )
        self.fc1 = nn.Linear(512, 1)

    def forward(self, image, label, **kwargs):
        x = image
        label = label.float()
        b, _, h, w = x.size()
        n = self.noise(x)
        r = self.rgb(x)
        n, r = self.ct(n, r)
        n = self.n_elayers(n)
        r = self.r_elayers(r)
        x = torch.cat((n, r), dim=1)
        x = F.adaptive_avg_pool2d(x, (1, 1)).view(x.size()[0], -1)
        x = self.fc1(x)
        x = x.squeeze(dim=1) if x.ndim == 2 and x.shape[1] == 1 else x

        combined_loss = F.binary_cross_entropy_with_logits(x, label)
        pred_label = x.sigmoid()

        dict = {
            "backward_loss": combined_loss,
            "pred_label": pred_label,
            "visual_loss": {
                "combined_loss": combined_loss
            }
        }
        return dict



from ForensicHub.core.base_model import BaseModel
from ForensicHub.registry import register_model

import torch
from transformers import AutoModelForImageTextToText, AutoProcessor

class Qwen3VLLLM(BaseModel):
    def __init__(self, backbone='Qwen/Qwen3-VL-8B-Instruct'):
        super(Qwen3VLLLM, self).__init__()

        assert 'qwen3_vl' in backbone, "Backbone must be one of Qwen3VL variants"
        self.model,self.processor = self.init_model(model_path=backbone)
    
    def init_model(self, model_path: str, **kwagrs):
        """Initialize the model from a pretrained checkpoint.
        
        Args:
            model_path (str): Path to the pretrained model checkpoint.
        """
        model = AutoModelForImageTextToText.from_pretrained(
            model_path,
            dtype=torch.bfloat16,
            attn_implementation="flash_attention_2",
            device_map="auto",
        )
        processor = AutoProcessor.from_pretrained(model_path)
        return model, processor

    def forward(self, image, max_new_tokens=1024, **kwargs):
        """Forward pass of the model.
        
        Args:
            image (torch.Tensor): Input image tensor.
            
        Returns:
            Dict[str, Any]: Dictionary containing model outputs.
                Must contain at least:
                    - 'pred': Model prediction
        """
        messages = self.build_messages(image)
        inputs = self.processor.apply_chat_template(
            messages,
            tokenize=True,
            add_generation_prompt=True,
            return_dict=True,
            return_tensors="pt"
        ).to(self.model.device)
        outputs = self.model.generate(**inputs, max_new_tokens=max_new_tokens, **kwargs)
        generated_ids_trimmed = [
            out_ids[len(in_ids) :] for in_ids, out_ids in zip(inputs.input_ids, outputs)
        ]
        output_text = self.processor.batch_decode(
            generated_ids_trimmed, skip_special_tokens=True, clean_up_tokenization_spaces=False
        )[0]
        return output_text

@register_model("Qwen3-VL-8B-Instruct")
class Qwen3VL8B(Qwen3VLLLM):
    def __init__(self):
        super().__init__(backbone='Qwen/Qwen3-VL-8B-Instruct')



import torch
import torchvision.datasets as datasets
from torchvision import transforms
from torch.utils.data import Dataset
import cv2
#import PIL.Image as Image
from PIL import Image, ImageOps, ImageEnhance
import numpy as np
import random
import os
import imutils

class MyDataset(Dataset):
    def __init__(self, txt, trans=None, option=None, type=None, size=None, augment=None):
        self.root = txt
        assert(type!=None)
        assert(size!=None)
        self.option = option
        if type == "DALLE":
            if option is not None:
                self.Neg_root = os.path.join("/data/xiziyi/DALLE/DA_"+ str(size) + "_JPG_postpro/DALLE/", option) + '/'
                self.Pos_root = os.path.join("/data/xiziyi/DALLE/DA_" + str(size) + "_JPG_postpro/ALASKA/", option) + '/'
            else:
                self.Neg_root = "/data/xiziyi/DALLE/DALLE_" + str(size) + "_JPG/"
                self.Pos_root = "/data/xiziyi/DALLE/ALASKA_" + str(size) + "_JPG/"
        elif type == "DreamStudio":
            if option is not None:
                self.Neg_root = os.path.join("/data/xiziyi/ds/DS_"+ str(size) +"_JPG_postpro/DS/", option) + '/'
                self.Pos_root = os.path.join("/data/xiziyi/ds/DS_"+ str(size) + "_JPG_postpro/ALASKA/", option) + '/'
            else:
                self.Neg_root = "/data/xiziyi/ds/DS_" + str(size) + "_JPG/"
                self.Pos_root = "/data/xiziyi/DALLE/ALASKA_"+ str(size) + "_JPG/"
        elif (type == "SPL2018" or type == 'DsTok') and option is not None:
            self.Neg_root = "/data/xiziyi/" + type + '/' + "split0/224_JPG_postpro/" + option + '/' + 'CG/'
            self.Pos_root = "/data/xiziyi/" + type + '/' + "split0/224_JPG_postpro/" + option + '/' + 'PG/'
            #print(self.Neg_root)
            #print(self.Pos_root)
        self.type = type
        self.image_list = self.get_dataset_info()
        self.augment = augment
        random.shuffle(self.image_list)
        self.tensor = transforms.ToTensor()
        if transforms:
            self.transforms = trans

    def __len__(self):
        return len(self.image_list)

    def __getitem__(self, item):
        img_path, label = self.image_list[item]
        image = Image.open(img_path).convert('RGB')
        #resize 256 or center crop
        '''
        if self.augment:
            method = random.randint(0, 7)
            #print(method)
            if method == 1:
                num = round(random.uniform(0.5, 2.5), 2)
                image = ImageEnhance.Color(image).enhance(num)
            elif method == 2:
                num = round(random.uniform(0.5, 2.5), 2)
                image = ImageEnhance.Brightness(image).enhance(num)
            elif method == 3:
                num = round(random.uniform(0.5, 2.5), 2)
                image = ImageEnhance.Contrast(image).enhance(num)
            elif method == 4:
                num = random.randint(0, 4)
                image = ImageEnhance.Sharpness(image).enhance(num)
            elif method == 5:
                img = cv2.cvtColor(np.asarray(image), cv2.COLOR_RGB2BGR)
                angle = random.randint(-180, 180)
                image = Image.fromarray(cv2.cvtColor(imutils.rotate(img, angle), cv2.COLOR_BGR2RGB))
            elif method == 6:
                img = cv2.cvtColor(np.asarray(image), cv2.COLOR_RGB2BGR)
                num = random.randint(0, 1)
                if num == 0:
                    img = cv2.GaussianBlur(img, (5, 5), 0)
                elif num == 1:
                    img = cv2.GaussianBlur(img, (3, 3), 0)
                image = Image.fromarray(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
            elif method == 7:
                img = cv2.cvtColor(np.asarray(image), cv2.COLOR_RGB2BGR)
                num = random.randint(0, 1)
                if num == 0:
                    img = cv2.blur(img, (5, 5), 0)
                elif num == 1:
                    img = cv2.blur(img, (3, 3), 0)
                image = Image.fromarray(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
        if self.type != "DALLE" or self.type != "DreamStudio":
            image = self.tensor(image)
        
        if image.width != 256 or image.height != 256:
            padding_w = 0
            padding_h = 0
            if image.width < 256:
                padding_w = ceil((256 - image.width) / 2)
            if image.height < 256:
                padding_h = ceil((256 - image.height) / 2)
            if padding_h != 0 or padding_w != 0:
                image = ImageOps.expand(image, (padding_w, padding_h, padding_w, padding_h))
            #image = cv2.copyMakeBorder(image, padding_h, padding_h, padding_w, padding_w,cv2.BORDER_REFLECT)
        startx = (image.height - 256) // 2
        starty = (image.width - 256) // 2
        image = image.crop((starty, startx, starty+256,startx+256))
        '''
        image = self.transforms(image)
        if label == "0":
            label = 0
        else:
            label = 1
        #torch.as_tensor(image)
        label = torch.as_tensor(label)
        return image, label

    def get_dataset_info(self):
        image_list = []
        with open(self.root,'r') as file:
            lines = file.read().splitlines()
        if (self.type == 'SPL2018' or self.type == 'DsTok') and (self.option is None):
            for line in lines:
                image_path = line[:-2]
                label = line[-1]
                image_list.append([image_path, label])
        else:
            #if self.type == 'DALLE' or self.type == 'DreamStudio':
            for line in lines:
                name = os.path.basename(line[:-2])
                #print(name)
                #if '.JPG' in name:
                #    name = name.replace('.JPG', '.jpg')
                # image_path.replace("openai_90","openai_75to85")
                label = line[-1]
                # 0: dalle
                if label == '0':
                    image_path = self.Neg_root + name
                # 1: alaska
                else:
                    image_path = self.Pos_root + name
                image_list.append([image_path, label])
        return image_list


#!/usr/bin/python3
# coding=utf-8
import timm
import numpy as np
import matplotlib.pyplot as plt
from efficientnet_pytorch import EfficientNet
import torch
import torch.nn as nn
import torch.nn.functional as F

from ForensicHub.registry import register_model
from ForensicHub.core.base_model import BaseModel

'''

'''


def weight_init(module):
    for n, m in module.named_children():
        # print('initialize: '+n)
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_normal_(m.weight, mode='fan_in', nonlinearity='relu')
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d)):
            nn.init.ones_(m.weight)
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, nn.Linear):
            nn.init.kaiming_normal_(m.weight, mode='fan_in', nonlinearity='relu')
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, nn.Sequential):
            weight_init(m)
        elif isinstance(m, nn.ReLU):
            pass
        elif isinstance(m, nn.Softmax):
            pass
        else:
            m.initialize()


class EfficientNetB3(nn.Module):
    def __init__(self):
        super(EfficientNetB3, self).__init__()
        self.inplanes = 64
        self.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=2, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.layer1 = self.make_layer(64, 3, stride=1, dilation=1)
        self.layer2 = self.make_layer(128, 4, stride=2, dilation=1)
        self.layer3 = self.make_layer(256, 6, stride=2, dilation=1)
        self.layer4 = self.make_layer(512, 3, stride=2, dilation=1)

    def make_layer(self, planes, blocks, stride, dilation):
        downsample = nn.Sequential(
            nn.Conv2d(self.inplanes, planes * 4, kernel_size=1, stride=stride, bias=False),
            nn.BatchNorm2d(planes * 4)
        )
        layers = [MBConvBlock(self.inplanes, planes, stride, downsample, dilation=dilation)]
        self.inplanes = planes * 4
        for _ in range(1, blocks):
            layers.append(MBConvBlock(self.inplanes, planes, dilation=dilation))
        return nn.Sequential(*layers)

    def forward(self, x):
        out1 = self.bn1(self.conv1(x))
        out1 = F.max_pool2d(F.relu(out1), kernel_size=3, stride=2, padding=1)
        out2 = self.layer1(out1)
        out3 = self.layer2(out2)
        out4 = self.layer3(out3)
        out5 = self.layer4(out4)
        return out2, out3, out4, out5

    def initialize(self):
        self.load_state_dict(torch.load('efficientnet-b3-5fb5a3c3.pth'), strict=False)


class MBConvBlock(nn.Module):
    def __init__(self, inplanes, planes, stride=1, downsample=None, dilation=1):
        super(MBConvBlock, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=(3 * dilation - 1) // 2,
                               bias=False, dilation=dilation)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.downsample = downsample

    def forward(self, x):
        identity = x
        out = F.relu(self.bn1(self.conv1(x)), inplace=True)
        out = F.relu(self.bn2(self.conv2(out)), inplace=True)
        out = self.bn3(self.conv3(out))
        if self.downsample is not None:
            identity = self.downsample(x)
        out += identity
        return F.relu(out, inplace=True)


class LHSE(nn.Module):
    def __init__(self):
        super(LHSE, self).__init__()
        self.conv1h = nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1)
        self.bn1h = nn.BatchNorm2d(64)
        self.conv2h = nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1)
        self.bn2h = nn.BatchNorm2d(64)
        self.conv3h = nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1)
        self.bn3h = nn.BatchNorm2d(64)
        self.conv4h = nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1)
        self.bn4h = nn.BatchNorm2d(64)

        self.conv1v = nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1)
        self.bn1v = nn.BatchNorm2d(64)
        self.conv2v = nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1)
        self.bn2v = nn.BatchNorm2d(64)
        self.conv3v = nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1)
        self.bn3v = nn.BatchNorm2d(64)
        self.conv4v = nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1)
        self.bn4v = nn.BatchNorm2d(64)

    def forward(self, left, down):
        if down.size()[2:] != left.size()[2:]:
            down = F.interpolate(down, size=left.size()[2:], mode='bilinear')
        out1h = F.relu(self.bn1h(self.conv1h(left)), inplace=True)
        out2h = F.relu(self.bn2h(self.conv2h(out1h)), inplace=True)
        out1v = F.relu(self.bn1v(self.conv1v(down)), inplace=True)
        out2v = F.relu(self.bn2v(self.conv2v(out1v)), inplace=True)
        fuse = out2h * out2v
        out3h = F.relu(self.bn3h(self.conv3h(fuse)), inplace=True) + out1h
        out4h = F.relu(self.bn4h(self.conv4h(out3h)), inplace=True)
        out3v = F.relu(self.bn3v(self.conv3v(fuse)), inplace=True) + out1v
        out4v = F.relu(self.bn4v(self.conv4v(out3v)), inplace=True)
        return out4h, out4v

    def initialize(self):
        weight_init(self)


class TFAM(nn.Module):
    def __init__(self, in_channels, out_channels, tfam_channels=64):
        super(TFAM, self).__init__()
        # Series of 1x1 conv to generate attention feature maps
        self.pab_channels = tfam_channels
        self.in_channels = in_channels
        self.top_conv = nn.Conv2d(in_channels, tfam_channels, kernel_size=1)
        self.center_conv = nn.Conv2d(in_channels, tfam_channels, kernel_size=1)
        self.bottom_conv = nn.Conv2d(in_channels, in_channels, kernel_size=3, padding=1)
        self.map_softmax = nn.Softmax(dim=1)
        self.out_conv = nn.Conv2d(in_channels, in_channels, kernel_size=3, padding=1)

    def forward(self, x):
        bsize = x.size()[0]
        h = x.size()[2]
        w = x.size()[3]
        x_top = self.top_conv(x)
        x_center = self.center_conv(x)
        x_bottom = self.bottom_conv(x)

        x_top = x_top.flatten(2)
        x_center = x_center.flatten(2).transpose(1, 2)
        x_bottom = x_bottom.flatten(2).transpose(1, 2)

        sp_map = torch.matmul(x_center, x_top)
        sp_map = self.map_softmax(sp_map.view(bsize, -1)).view(bsize, h * w, h * w)
        sp_map = torch.matmul(sp_map, x_bottom)
        sp_map = sp_map.reshape(bsize, self.in_channels, h, w)
        x = x + sp_map
        x = self.out_conv(x)
        return x

    def initialize(self):
        weight_init(self)


class Decoder(nn.Module):
    def __init__(self):
        super(Decoder, self).__init__()
        self.lhse45 = LHSE()
        self.lhse34 = LHSE()
        self.lhse23 = LHSE()
        self.tfam = TFAM(64, 64, tfam_channels=16)  # 如果显存溢出可减小tfam_channels

    def forward(self, out2h, out3h, out4h, out5v, fback=None):
        if fback is not None:
            refine5 = F.interpolate(fback, size=out5v.size()[2:], mode='bilinear')
            refine4 = F.interpolate(fback, size=out4h.size()[2:], mode='bilinear')
            refine3 = F.interpolate(fback, size=out3h.size()[2:], mode='bilinear')
            refine2 = F.interpolate(fback, size=out2h.size()[2:], mode='bilinear')
            out5v = out5v + refine5
            out4h, out4v = self.lhse45(out4h + refine4, out5v)
            out3h, out3v = self.lhse34(out3h + refine3, out4v)
            out2h, pred = self.lhse23(out2h + refine2, out3v)
        else:
            out4h, out4v = self.lhse45(out4h, out5v)
            out3h, out3v = self.lhse34(out3h, out4v)
            out2h, pred = self.lhse23(out2h, out3v)

        pred = self.tfam(pred)
        return pred

    def initialize(self):
        weight_init(self)


@register_model("Tifdm")
class Tifdm(BaseModel):
    def __init__(self, backbone_model_name='resnet50'):
        super(Tifdm, self).__init__()
        self.backbone = timm.create_model(backbone_model_name, features_only=True, pretrained=True)
        self.squeeze5 = nn.Sequential(nn.Conv2d(2048, 64, 1), nn.BatchNorm2d(64), nn.ReLU(inplace=True))
        self.squeeze4 = nn.Sequential(nn.Conv2d(1024, 64, 1), nn.BatchNorm2d(64), nn.ReLU(inplace=True))
        self.squeeze3 = nn.Sequential(nn.Conv2d(512, 64, 1), nn.BatchNorm2d(64), nn.ReLU(inplace=True))
        self.squeeze2 = nn.Sequential(nn.Conv2d(256, 64, 1), nn.BatchNorm2d(64), nn.ReLU(inplace=True))

        self.decoder = Decoder()
        self.linearp = nn.Conv2d(64, 2, kernel_size=3, stride=1, padding=1)
        self.initialize()

    def cal_seg_loss(self, pred, gt):
        h, w = gt.shape[-2:]
        pred = F.interpolate(pred, size=(h, w), mode='bilinear')
        ce_loss = F.cross_entropy(pred, gt)
        return ce_loss, pred

    def forward(self, image, mask, **kwargs):
        x = image
        mask = mask.squeeze(1).long()  # [B,1,H,W] -> [B,H,W]
        out2h, out3h, out4h, out5v = self.backbone(x)[1:]
        out2h, out3h, out4h, out5v = self.squeeze2(out2h), self.squeeze3(out3h), self.squeeze4(out4h), self.squeeze5(
            out5v)
        pred1 = self.decoder(out2h, out3h, out4h, out5v)
        output = self.linearp(pred1)
        seg_loss, output = self.cal_seg_loss(output, mask)
        output = F.softmax(output,dim=1)[:,1:]
        output_dict = {
            "backward_loss": seg_loss,
            "pred_mask": output,
            "visual_loss": {
                "seg_loss": seg_loss,
                "combined_loss": seg_loss
            },
            "visual_image": {
                "pred_mask": output,
            }
        }
        return output_dict

    def initialize(self):
        weight_init(self.decoder)


if __name__ == "__main__":
    model = Tifdm()
    img = torch.ones((1, 3, 64, 64))
    mask = torch.ones((1, 1, 64, 64), dtype=torch.int64)
    pred = model(img, mask)
    print(pred)



#!/usr/bin/python3
# coding=utf-8
import timm
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

BN_MOMENTUM = 0.01

from ForensicHub.registry import register_model
from ForensicHub.core.base_model import BaseModel

'''
Progressive Supervision for Tampering Localization in Document Images
https://link.springer.com/chapter/10.1007/978-981-99-8184-7_11
'''


class NonLocalMask(nn.Module):
    def __init__(self, in_channels, reduce_scale):
        super(NonLocalMask, self).__init__()

        self.r = reduce_scale

        # input channel number
        self.ic = in_channels * self.r * self.r

        # middle channel number
        self.mc = self.ic

        self.g = nn.Conv2d(in_channels=self.ic, out_channels=self.ic,
                           kernel_size=1, stride=1, padding=0)

        self.theta = nn.Conv2d(in_channels=self.ic, out_channels=self.mc,
                               kernel_size=1, stride=1, padding=0)
        self.phi = nn.Conv2d(in_channels=self.ic, out_channels=self.mc,
                             kernel_size=1, stride=1, padding=0)

        self.W_s = nn.Conv2d(in_channels=in_channels, out_channels=in_channels,
                             kernel_size=1, stride=1, padding=0)

        self.W_c = nn.Conv2d(in_channels=in_channels, out_channels=in_channels,
                             kernel_size=1, stride=1, padding=0)

        self.gamma_s = nn.Parameter(torch.ones(1))

        self.gamma_c = nn.Parameter(torch.ones(1))

        self.getmask = nn.Sequential(
            nn.Conv2d(in_channels=in_channels, out_channels=16, kernel_size=3, stride=1, padding=1),
            nn.ReLU(),
            nn.Conv2d(in_channels=16, out_channels=1, kernel_size=3, stride=1, padding=1))

    def forward(self, x):
        """
            inputs :
                x : input feature maps( B X C X H X W)
            value :
                f: B X (HxW) X (HxW)
                ic: intermediate channels
                z: feature maps( B X C X H X W)
            output:
                mask: feature maps( B X 1 X H X W)
        """

        b, c, h, w = x.shape

        x1 = x.reshape(b, self.ic, h // self.r, w // self.r)

        # g x
        g_x = self.g(x1).view(b, self.ic, -1)
        g_x = g_x.permute(0, 2, 1)

        # theta
        theta_x = self.theta(x1).view(b, self.mc, -1)

        theta_x_s = theta_x.permute(0, 2, 1)
        theta_x_c = theta_x

        # phi x
        phi_x = self.phi(x1).view(b, self.mc, -1)

        phi_x_s = phi_x
        phi_x_c = phi_x.permute(0, 2, 1)

        # non-local attention
        f_s = torch.matmul(theta_x_s, phi_x_s)
        f_s_div = F.softmax(f_s, dim=-1)

        f_c = torch.matmul(theta_x_c, phi_x_c)
        f_c_div = F.softmax(f_c, dim=-1)

        # get y_s
        y_s = torch.matmul(f_s_div, g_x)
        y_s = y_s.permute(0, 2, 1).contiguous()
        y_s = y_s.view(b, c, h, w)

        # get y_c
        y_c = torch.matmul(g_x, f_c_div)
        y_c = y_c.view(b, c, h, w)

        # get z
        z = x + self.gamma_s * self.W_s(y_s) + self.gamma_c * self.W_c(y_c)

        # get mask
        mask = torch.sigmoid(self.getmask(z.clone()))

        return mask, z


class NLCDetection(nn.Module):
    def __init__(self, ):
        super(NLCDetection, self).__init__()

        feat1_num, feat2_num, feat3_num, feat4_num = 64, 128, 256, 512

        self.getmask4 = NonLocalMask(feat4_num, 1)
        self.getmask3 = NonLocalMask(feat3_num, 2)
        self.getmask2 = NonLocalMask(feat2_num, 2)
        self.getmask1 = NonLocalMask(feat1_num, 4)

    def forward(self, feat):
        """
            inputs :
                feat : a list contains features from s1, s2, s3, s4
            output:
                mask1: output mask ( B X 1 X H X W)
                pred_cls: output cls (B X 4)
        """
        s1, s2, s3, s4 = feat

        mask4, z4 = self.getmask4(s4)
        mask4U = F.interpolate(mask4, size=s3.size()[2:], mode='bilinear', align_corners=True)

        s3 = s3 * mask4U
        mask3, z3 = self.getmask3(s3)
        mask3U = F.interpolate(mask3, size=s2.size()[2:], mode='bilinear', align_corners=True)

        s2 = s2 * mask3U
        mask2, z2 = self.getmask2(s2)
        mask2U = F.interpolate(mask2, size=s1.size()[2:], mode='bilinear', align_corners=True)

        s1 = s1 * mask2U
        mask1, z1 = self.getmask1(s1)

        return [mask1, mask2, mask3, mask4]


class Bottleneck(nn.Module):
    expansion = 2

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.conv3 = nn.Conv2d(planes, planes * self.expansion, kernel_size=1,
                               bias=False)
        self.bn3 = nn.BatchNorm2d(planes * self.expansion,
                                  momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class DetectionHead(nn.Module):
    def __init__(self, ):
        super(DetectionHead, self).__init__()

        pre_stage_channels = (64, 128, 256, 512)

        # classification head
        self.incre_modules, self.downsamp_modules, \
            self.final_layer = self._make_head(pre_stage_channels)

        self.classifier = nn.Sequential(
            nn.Linear(128, 16),
            nn.ReLU(inplace=True),
            nn.Linear(16, 1)
        )

    def _make_layer(self, block, inplanes, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion, momentum=BN_MOMENTUM),
            )

        layers = []
        layers.append(block(inplanes, planes, stride, downsample))
        inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(inplanes, planes))

        return nn.Sequential(*layers)

    def _make_head(self, pre_stage_channels):
        head_block = Bottleneck
        head_channels = pre_stage_channels

        incre_modules = []
        for i, channels in enumerate(pre_stage_channels):
            incre_module = self._make_layer(head_block,
                                            channels,
                                            head_channels[i],
                                            1,
                                            stride=1)
            incre_modules.append(incre_module)
        incre_modules = nn.ModuleList(incre_modules)

        # downsampling modules
        downsamp_modules = []
        for i in range(len(pre_stage_channels) - 1):
            in_channels = head_channels[i] * head_block.expansion
            out_channels = head_channels[i + 1] * head_block.expansion

            downsamp_module = nn.Sequential(
                nn.Conv2d(in_channels=in_channels,
                          out_channels=out_channels,
                          kernel_size=3,
                          stride=2,
                          padding=1),
                nn.BatchNorm2d(out_channels, momentum=BN_MOMENTUM),
                nn.ReLU(inplace=True)
            )

            downsamp_modules.append(downsamp_module)
        downsamp_modules = nn.ModuleList(downsamp_modules)

        final_layer = nn.Sequential(
            nn.Conv2d(
                in_channels=head_channels[3] * head_block.expansion,
                out_channels=128,
                kernel_size=1,
                stride=1,
                padding=0
            ),
            nn.BatchNorm2d(128, momentum=BN_MOMENTUM),
            nn.ReLU(inplace=True)
        )

        return incre_modules, downsamp_modules, final_layer

    def forward(self, y_list):

        y = self.incre_modules[0](y_list[0])
        for i in range(len(self.downsamp_modules)):
            y = self.incre_modules[i + 1](y_list[i + 1]) + \
                self.downsamp_modules[i](y)

        y = self.final_layer(y)

        # average and flatten
        y = F.avg_pool2d(y, kernel_size=y.size()[2:]).view(y.size(0), -1)

        logit = self.classifier(y).squeeze(1)

        return logit


class BayerConv(nn.Module):
    def __init__(self, in_channel=3):
        super(BayerConv, self).__init__()
        self.BayarConv2D = nn.Conv2d(in_channel, 3, 5, 1, padding=2, bias=False)

        # Convert tensors to Parameters
        bayar_mask = np.ones(shape=(5, 5))
        bayar_mask[2, 2] = 0
        self.bayar_mask = nn.Parameter(torch.tensor(bayar_mask, dtype=torch.float32), requires_grad=False)

        bayar_final = np.zeros((5, 5))
        bayar_final[2, 2] = -1
        self.bayar_final = nn.Parameter(torch.tensor(bayar_final, dtype=torch.float32), requires_grad=False)

    def forward(self, x):
        self.BayarConv2D.weight.data *= self.bayar_mask
        self.BayarConv2D.weight.data *= torch.pow(self.BayarConv2D.weight.data.sum(axis=(2, 3)).view(3, 3, 1, 1), -1)
        self.BayarConv2D.weight.data += self.bayar_final

        return self.BayarConv2D(x)


def lbp_torch(x):
    # pad image for 3x3 mask size
    x = F.pad(input=x, pad=[1, 1, 1, 1], mode='constant')
    b = x.shape
    M = b[1]
    N = b[2]

    y = x
    # select elements within 3x3 mask
    y00 = y[:, 0:M - 2, 0:N - 2]
    y01 = y[:, 0:M - 2, 1:N - 1]
    y02 = y[:, 0:M - 2, 2:N]
    #
    y10 = y[:, 1:M - 1, 0:N - 2]
    y11 = y[:, 1:M - 1, 1:N - 1]
    y12 = y[:, 1:M - 1, 2:N]
    #
    y20 = y[:, 2:M, 0:N - 2]
    y21 = y[:, 2:M, 1:N - 1]
    y22 = y[:, 2:M, 2:N]

    # Apply comparisons and multiplications
    bit = torch.ge(y01, y11)
    tmp = torch.mul(bit, torch.tensor(1))

    bit = torch.ge(y02, y11)
    val = torch.mul(bit, torch.tensor(2))
    val = torch.add(val, tmp)

    bit = torch.ge(y12, y11)
    tmp = torch.mul(bit, torch.tensor(4))
    val = torch.add(val, tmp)

    bit = torch.ge(y22, y11)
    tmp = torch.mul(bit, torch.tensor(8))
    val = torch.add(val, tmp)

    bit = torch.ge(y21, y11)
    tmp = torch.mul(bit, torch.tensor(16))
    val = torch.add(val, tmp)

    bit = torch.ge(y20, y11)
    tmp = torch.mul(bit, torch.tensor(32))
    val = torch.add(val, tmp)

    bit = torch.ge(y10, y11)
    tmp = torch.mul(bit, torch.tensor(64))
    val = torch.add(val, tmp)

    bit = torch.ge(y00, y11)
    tmp = torch.mul(bit, torch.tensor(128))
    val = torch.add(val, tmp)
    return val[..., 1:-1]


@register_model("PSNet")
class PSNet(BaseModel):
    def __init__(self, ):
        super(PSNet, self).__init__()
        self.bayar = BayerConv()
        self.backbone = timm.create_model('hrnet_w18', features_only=True, pretrained=True)
        self.backbone.conv1 = nn.Conv2d(7, 64, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False)
        self.conv = nn.Conv2d(1024, 512, 1, 1, 0)
        self.cls = DetectionHead()
        self.upsample = nn.UpsamplingBilinear2d(scale_factor=2.0)
        self.decoder = NLCDetection()
        self.kl_loss = torch.nn.KLDivLoss()
        self.maxp = nn.AdaptiveMaxPool2d(1)

    def cal_seg_loss(self, pred, gt):
        h, w = gt.shape[-2:]
        pred = F.interpolate(pred, size=(h, w), mode='bilinear')
        ce_loss = F.binary_cross_entropy_with_logits(pred, gt)
        return ce_loss, pred

    def forward(self, image, mask, **kwargs):
        x = image
        lbp = lbp_torch(x)
        bayar = (x - self.bayar(x))
        bayer = torch.where(bayar > 0, x, 1 - x)
        x = torch.cat((x, bayar, lbp), 1)
        feats = self.backbone(x)
        feats[-2] = (feats[-2] + self.upsample(self.conv(feats[-1])))
        del feats[-1]
        output = self.decoder(feats)
        cls_output = self.cls(feats)
        seg_loss, seg_pred = self.cal_seg_loss(output[0], mask)
        output[0] = seg_pred
        for i in range(3):
            seg_loss = (seg_loss + self.cal_seg_loss(output[i + 1], mask)[0])
        kl_loss = F.kl_div(output[0].flatten(1).max(1).values, cls_output)
        cls_loss = F.binary_cross_entropy_with_logits(cls_output, mask.flatten(1).max(1).values)
        combined_loss = (seg_loss + cls_loss + kl_loss)
        output_dict = {
            "backward_loss": combined_loss,
            "pred_mask": seg_pred,
            "pred_label": cls_output,
            "visual_loss": {
                "seg_loss": seg_loss,
                "cls_loss": cls_loss,
                "kl_loss": kl_loss,
                "combined_loss": combined_loss
            },
            "visual_image": {
                "pred_mask": seg_pred,
            }
        }
        return output_dict


if __name__ == "__main__":
    model = PSNet()
    img = torch.ones((1, 3, 64, 64))
    mask = torch.ones((1, 1, 64, 64), dtype=torch.float32)
    pred = model(img, mask)  # pred[0] for final segmentation prediction
    print(pred)



import torch
import jpegio
import numpy as np
import torch.nn as nn
import torch.optim as optim
import torch.distributed as dist
from torch.autograd import Variable
from .fph import FPH
from .dwt import DWTFPN
from .backbone_convnext import ConvNeXt
import torch.nn.functional as F
from timm.models.layers import trunc_normal_, DropPath
from functools import partial
from typing import Optional, Union, List, Dict, Any

from ForensicHub.registry import register_model
from ForensicHub.core.base_model import BaseModel

'''

'''


class LayerNorm(nn.Module):
    def __init__(self, normalized_shape, eps=1e-6, data_format="channels_last"):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.eps = eps
        self.data_format = data_format
        if self.data_format not in ["channels_last", "channels_first"]:
            raise NotImplementedError
        self.normalized_shape = (normalized_shape,)

    def forward(self, x):
        if self.data_format == "channels_last":
            return F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        elif self.data_format == "channels_first":
            u = x.mean(1, keepdim=True)
            s = (x - u).pow(2).mean(1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.eps)
            x = self.weight[:, None, None] * x + self.bias[:, None, None]
            return x


def get_norm_layer(
        use_norm: Union[bool, str, Dict[str, Any]], out_channels: int
) -> nn.Module:
    supported_norms = ("inplace", "batchnorm", "identity", "layernorm", "instancenorm")

    # Step 1. Convert tot dict representation

    ## Check boolean
    if use_norm is True:
        norm_params = {"type": "batchnorm"}
    elif use_norm is False:
        norm_params = {"type": "identity"}

    ## Check string
    elif isinstance(use_norm, str):
        norm_str = use_norm.lower()
        if norm_str == "inplace":
            norm_params = {
                "type": "inplace",
                "activation": "leaky_relu",
                "activation_param": 0.0,
            }
        elif norm_str in supported_norms:
            norm_params = {"type": norm_str}
        else:
            raise ValueError(
                f"Unrecognized normalization type string provided: {use_norm}. Should be in "
                f"{supported_norms}"
            )

    ## Check dict
    elif isinstance(use_norm, dict):
        norm_params = use_norm

    else:
        raise ValueError(
            f"Invalid type for use_norm should either be a bool (batchnorm/identity), "
            f"a string in {supported_norms}, or a dict like {{'type': 'batchnorm', **kwargs}}"
        )

    # Step 2. Check if the dict is valid
    if "type" not in norm_params:
        raise ValueError(
            f"Malformed dictionary given in use_norm: {use_norm}. Should contain key 'type'."
        )
    if norm_params["type"] not in supported_norms:
        raise ValueError(
            f"Unrecognized normalization type string provided: {use_norm}. Should be in {supported_norms}"
        )
    if norm_params["type"] == "inplace" and InPlaceABN is None:
        raise RuntimeError(
            "In order to use `use_norm='inplace'` the inplace_abn package must be installed. Use:\n"
            "  $ pip install -U wheel setuptools\n"
            "  $ pip install inplace_abn --no-build-isolation\n"
            "Also see: https://github.com/mapillary/inplace_abn"
        )

    # Step 3. Initialize the norm layer
    norm_type = norm_params["type"]
    norm_kwargs = {k: v for k, v in norm_params.items() if k != "type"}

    if norm_type == "inplace":
        norm = InPlaceABN(out_channels, **norm_kwargs)
    elif norm_type == "batchnorm":
        norm = nn.BatchNorm2d(out_channels, **norm_kwargs)
    elif norm_type == "identity":
        norm = nn.Identity()
    elif norm_type == "layernorm":
        norm = nn.LayerNorm(out_channels, **norm_kwargs)
    elif norm_type == "instancenorm":
        norm = nn.InstanceNorm2d(out_channels, **norm_kwargs)
    else:
        raise ValueError(f"Unrecognized normalization type: {norm_type}")

    return norm


class Conv2dReLU(nn.Sequential):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: int,
            padding: int = 0,
            stride: int = 1,
            use_batchnorm: bool = True,
    ):
        if use_batchnorm:
            norm = get_norm_layer(use_batchnorm, out_channels)

        is_identity = isinstance(norm, nn.Identity)
        conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            bias=is_identity,
        )

        activation = nn.ReLU(inplace=True)

        super(Conv2dReLU, self).__init__(conv, norm, activation)


class SCSEModule(nn.Module):
    def __init__(self, in_channels, reduction=16):
        super().__init__()
        self.cSE = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_channels, in_channels // reduction, 1),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels // reduction, in_channels, 1),
            nn.Sigmoid(),
        )
        self.sSE = nn.Sequential(nn.Conv2d(in_channels, 1, 1), nn.Sigmoid())

    def forward(self, x):
        return x * self.cSE(x) + x * self.sSE(x)


class ConvBlock(nn.Module):
    def __init__(self, dim, drop_path=0., layer_scale_init_value=1e-6):
        super().__init__()
        self.dwconv = nn.Conv2d(dim, dim, kernel_size=7, padding=3, groups=dim)
        self.norm = LayerNorm(dim, eps=1e-6)
        self.pwconv1 = nn.Linear(dim, 4 * dim)
        self.act = nn.GELU()
        self.pwconv2 = nn.Linear(4 * dim, dim)
        self.gamma = nn.Parameter(layer_scale_init_value * torch.ones((dim)),
                                  requires_grad=True) if layer_scale_init_value > 0 else None
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

    def forward(self, x):
        ipt = x
        x = self.dwconv(x)
        x = x.permute(0, 2, 3, 1)
        x = self.norm(x)
        x = self.pwconv1(x)
        x = self.act(x)
        x = self.pwconv2(x)
        if self.gamma is not None:
            x = self.gamma * x
        x = x.permute(0, 3, 1, 2)
        x = ipt + self.drop_path(x)
        return x


class AddCoords(nn.Module):
    def __init__(self, with_r=True):
        super().__init__()
        self.with_r = with_r

    def forward(self, input_tensor):
        batch_size, _, x_dim, y_dim = input_tensor.size()
        xx_c, yy_c = torch.meshgrid(torch.arange(x_dim, dtype=input_tensor.dtype),
                                    torch.arange(y_dim, dtype=input_tensor.dtype))
        xx_c = xx_c.to(input_tensor.device) / (x_dim - 1) * 2 - 1
        yy_c = yy_c.to(input_tensor.device) / (y_dim - 1) * 2 - 1
        xx_c = xx_c.expand(batch_size, 1, x_dim, y_dim)
        yy_c = yy_c.expand(batch_size, 1, x_dim, y_dim)
        ret = torch.cat((input_tensor, xx_c, yy_c), dim=1)
        if self.with_r:
            rr = torch.sqrt(torch.pow(xx_c - 0.5, 2) + torch.pow(yy_c - 0.5, 2))
            ret = torch.cat([ret, rr], dim=1)
        return ret


class VPH(nn.Module):
    def __init__(self, dims=[96, 192], drop_path_rate=0.4, layer_scale_init_value=1e-6):
        super().__init__()
        dp_rates = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]
        self.downsample_layers = nn.ModuleList([nn.Sequential(nn.Conv2d(6, dims[0], kernel_size=4, stride=4),
                                                              LayerNorm(dims[0], eps=1e-6,
                                                                        data_format="channels_first")), nn.Sequential(
            LayerNorm(dims[1], eps=1e-6, data_format="channels_first"),
            nn.Conv2d(dims[1], dims[2], kernel_size=2, stride=2))])
        self.stages = nn.ModuleList([nn.Sequential(
            *[ConvBlock(dim=dims[0], drop_path=dp_rates[j], layer_scale_init_value=layer_scale_init_value) for j in
              range(3)]), nn.Sequential(
            *[ConvBlock(dim=dims[1], drop_path=dp_rates[3 + j], layer_scale_init_value=layer_scale_init_value) for j in
              range(3)])])
        self.apply(self._init_weights)

    def initnorm(self):
        norm_layer = partial(LayerNorm, eps=1e-6, data_format="channels_first")
        for i_layer in range(4):
            layer = norm_layer(self.dims[i_layer])
            layer_name = f'norm{i_layer}'
            self.add_module(layer_name, layer)

    def _init_weights(self, m):
        if isinstance(m, (nn.Conv2d, nn.Linear)):
            trunc_normal_(m.weight, std=.02)
            nn.init.constant_(m.bias, 0)

    def init_weights(self, pretrained=None):
        def _init_weights(m):
            if isinstance(m, nn.Linear):
                trunc_normal_(m.weight, std=.02)
                if isinstance(m, nn.Linear) and m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

        self.apply(_init_weights)

    def forward(self, x):
        outs = []
        x = self.stages[0](self.downsample_layers[0](x))
        outs = [self.norm0(x)]
        x = self.stages[1](self.downsample_layers[1](x))
        outs.append(self.norm1(x))
        return outs


class DecoderBlock(nn.Module):
    def __init__(self, cin, cadd, cout, ):
        super().__init__()
        self.cin = (cin + cadd)
        self.cout = cout
        self.conv1 = Conv2dReLU(self.cin, self.cout, kernel_size=3, padding=1, use_batchnorm=True)
        self.conv2 = Conv2dReLU(self.cout, self.cout, kernel_size=3, padding=1, use_batchnorm=True)

    def forward(self, x1, x2=None):
        x1 = F.interpolate(x1, scale_factor=2.0, mode="nearest")
        if x2 is not None:
            x1 = torch.cat([x1, x2], dim=1)
        x1 = self.conv1(x1[:, :self.cin])
        x1 = self.conv2(x1)
        return x1


class ConvBNReLU(nn.Module):
    def __init__(self, in_c, out_c, ks, stride=1, norm=True, res=False):
        super(ConvBNReLU, self).__init__()
        if norm:
            self.conv = nn.Sequential(
                nn.Conv2d(in_c, out_c, kernel_size=ks, padding=ks // 2, stride=stride, bias=False),
                nn.BatchNorm2d(out_c), nn.ReLU(True))
        else:
            self.conv = nn.Conv2d(in_c, out_c, kernel_size=ks, padding=ks // 2, stride=stride, bias=False)
        self.res = res

    def forward(self, x):
        if self.res:
            return (x + self.conv(x))
        else:
            return self.conv(x)


class FUSE1(nn.Module):
    def __init__(self, in_channels_list=(96, 192, 384, 768)):
        super(FUSE1, self).__init__()
        self.c31 = ConvBNReLU(in_channels_list[2], in_channels_list[2], 1)
        self.c32 = ConvBNReLU(in_channels_list[3], in_channels_list[2], 1)
        self.c33 = ConvBNReLU(in_channels_list[2], in_channels_list[2], 3)

        self.c21 = ConvBNReLU(in_channels_list[1], in_channels_list[1], 1)
        self.c22 = ConvBNReLU(in_channels_list[2], in_channels_list[1], 1)
        self.c23 = ConvBNReLU(in_channels_list[1], in_channels_list[1], 3)

        self.c11 = ConvBNReLU(in_channels_list[0], in_channels_list[0], 1)
        self.c12 = ConvBNReLU(in_channels_list[1], in_channels_list[0], 1)
        self.c13 = ConvBNReLU(in_channels_list[0], in_channels_list[0], 3)

    def forward(self, x):
        x, x1, x2, x3 = x
        h, w = x2.shape[-2:]
        x2 = self.c33(F.interpolate(self.c32(x3), size=(h, w)) + self.c31(x2))
        h, w = x1.shape[-2:]
        x1 = self.c23(F.interpolate(self.c22(x2), size=(h, w)) + self.c21(x1))
        h, w = x.shape[-2:]
        x = self.c13(F.interpolate(self.c12(x1), size=(h, w)) + self.c11(x))
        return x, x1, x2, x3


class FUSE2(nn.Module):
    def __init__(self, in_channels_list=(96, 192, 384)):
        super(FUSE2, self).__init__()

        self.c21 = ConvBNReLU(in_channels_list[1], in_channels_list[1], 1)
        self.c22 = ConvBNReLU(in_channels_list[2], in_channels_list[1], 1)
        self.c23 = ConvBNReLU(in_channels_list[1], in_channels_list[1], 3)

        self.c11 = ConvBNReLU(in_channels_list[0], in_channels_list[0], 1)
        self.c12 = ConvBNReLU(in_channels_list[1], in_channels_list[0], 1)
        self.c13 = ConvBNReLU(in_channels_list[0], in_channels_list[0], 3)

    def forward(self, x):
        x, x1, x2 = x
        h, w = x1.shape[-2:]
        x1 = self.c23(F.interpolate(self.c22(x2), size=(h, w), mode='bilinear', align_corners=True) + self.c21(x1))
        h, w = x.shape[-2:]
        x = self.c13(F.interpolate(self.c12(x1), size=(h, w), mode='bilinear', align_corners=True) + self.c11(x))
        return x, x1, x2


class FUSE3(nn.Module):
    def __init__(self, in_channels_list=(96, 192)):
        super(FUSE3, self).__init__()

        self.c11 = ConvBNReLU(in_channels_list[0], in_channels_list[0], 1)
        self.c12 = ConvBNReLU(in_channels_list[1], in_channels_list[0], 1)
        self.c13 = ConvBNReLU(in_channels_list[0], in_channels_list[0], 3)

    def forward(self, x):
        x, x1 = x
        h, w = x.shape[-2:]
        x = self.c13(F.interpolate(self.c12(x1), size=(h, w), mode='bilinear', align_corners=True) + self.c11(x))
        return x, x1


class MID(nn.Module):
    def __init__(self, encoder_channels, decoder_channels):
        super().__init__()
        encoder_channels = encoder_channels[1:][::-1]
        self.in_channels = [encoder_channels[0]] + list(decoder_channels[:-1])
        self.add_channels = list(encoder_channels[1:]) + [96]
        self.out_channels = decoder_channels
        self.fuse1 = FUSE1()
        self.fuse2 = FUSE2()
        self.fuse3 = FUSE3()
        decoder_convs = {}
        for layer_idx in range(len(self.in_channels) - 1):
            for depth_idx in range(layer_idx + 1):
                if depth_idx == 0:
                    in_ch = self.in_channels[layer_idx]
                    skip_ch = self.add_channels[layer_idx] * (layer_idx + 1)
                    out_ch = self.out_channels[layer_idx]
                else:
                    out_ch = self.add_channels[layer_idx]
                    skip_ch = self.add_channels[layer_idx] * (layer_idx + 1 - depth_idx)
                    in_ch = self.add_channels[layer_idx - 1]
                decoder_convs[f"x_{depth_idx}_{layer_idx}"] = DecoderBlock(in_ch, skip_ch, out_ch)
        decoder_convs[f"x_{0}_{len(self.in_channels) - 1}"] = DecoderBlock(self.in_channels[-1], 0,
                                                                           self.out_channels[-1])
        self.decoder_convs = nn.ModuleDict(decoder_convs)

    def forward(self, *features):
        decoder_features = {}
        features = self.fuse1(features)[::-1]
        decoder_features["x_0_0"] = self.decoder_convs["x_0_0"](features[0], features[1])
        decoder_features["x_1_1"] = self.decoder_convs["x_1_1"](features[1], features[2])
        decoder_features["x_2_2"] = self.decoder_convs["x_2_2"](features[2], features[3])
        decoder_features["x_2_2"], decoder_features["x_1_1"], decoder_features["x_0_0"] = self.fuse2(
            (decoder_features["x_2_2"], decoder_features["x_1_1"], decoder_features["x_0_0"]))
        decoder_features["x_0_1"] = self.decoder_convs["x_0_1"](decoder_features["x_0_0"],
                                                                torch.cat((decoder_features["x_1_1"], features[2]), 1))
        decoder_features["x_1_2"] = self.decoder_convs["x_1_2"](decoder_features["x_1_1"],
                                                                torch.cat((decoder_features["x_2_2"], features[3]), 1))
        decoder_features["x_1_2"], decoder_features["x_0_1"] = self.fuse3(
            (decoder_features["x_1_2"], decoder_features["x_0_1"]))
        decoder_features["x_0_2"] = self.decoder_convs["x_0_2"](decoder_features["x_0_1"], torch.cat(
            (decoder_features["x_1_2"], decoder_features["x_2_2"], features[3]), 1))
        return self.decoder_convs["x_0_3"](
            torch.cat((decoder_features["x_0_2"], decoder_features["x_1_2"], decoder_features["x_2_2"]), 1))


@register_model("FFDN")
class FFDN(BaseModel):
    def __init__(self, decoder_channels=(384, 192, 96, 64), classes=2,
                 weight_path='/mnt/data1/public_datasets/Doc/Hub/ffdn/convnext_small.pth'):
        super().__init__()
        self.vph = ConvNeXt()  # Backbone1, using stage 0, 1
        self.fph = FPH()
        self.decoder = MID(encoder_channels=(96, 192, 384, 768), decoder_channels=decoder_channels)
        self.head = nn.Sequential(nn.Conv2d(256, 256, 3, 1, 1), nn.ReLU(), nn.Conv2d(256, 2, 1, 1, 0))
        self.addcoords = AddCoords()
        self.FU = nn.Sequential(SCSEModule(448), nn.Conv2d(448, 192, 3, 1, 1), nn.BatchNorm2d(192), nn.ReLU(True),
                                nn.Conv2d(192, 192, 1, 1, 0))
        self.FU[-1].weight.data.zero_()
        self.decoder = DWTFPN([96, 192, 384, 768], 256)
        self.init_vph(weight_path)

    def init_vph(self, weight_path):
        weights = torch.load(weight_path)['state_dict']
        dels = [k for k in weights.keys() if not k.startswith('backbone.')]
        for k in dels:
            del weights[k]
        new_weights = {}
        for k, v in weights.items():
            new_weights['.'.join(k.split('.')[1:])] = v
        self.vph.load_state_dict(new_weights)

    def cal_seg_loss(self, pred, gt):
        h, w = gt.shape[-2:]
        pred = F.interpolate(pred, size=(h, w), mode='bilinear')
        ce_loss = F.cross_entropy(pred, gt)
        return ce_loss, pred

    def forward(self, image, dct, qt, mask, **kwargs):
        DCT_coef = dct
        qtables = qt
        x = image
        mask = mask.squeeze(1).long()  # [B,1,H,W] -> [B,H,W]
        DCT_coef = DCT_coef.squeeze(1).long()  # [B,1,H,W] -> [B,H,W]
        if len(qtables.shape) == 3:
            qtables = qtables.unsqueeze(1)
        features = self.vph.forward_features(x, end_index=2)
        features[1] = self.FU(torch.cat((features[1], self.fph(DCT_coef, qtables)), 1)) + features[1]
        features.extend(self.vph.forward_features(features[1], start_index=2, end_index=4))
        decoder_output = self.decoder(features)
        output = self.head(decoder_output[0])
        seg_loss, output = self.cal_seg_loss(output, mask)
        output = F.softmax(output, dim=1)   
        output = output[:, 1:]
        output_dict = {
            "backward_loss": seg_loss,
            "pred_mask": output,
            "visual_loss": {
                "seg_loss": seg_loss,
            },
            "visual_image": {
                "pred_mask": output,
            }
        }
        return output_dict


if __name__ == "__main__":
    img = torch.ones((1, 3, 512, 512))
    mask = torch.ones((1, 1, 512, 512), dtype=torch.int64)
    dct = torch.ones((1, 1, 512, 512), dtype=torch.int64)  # From JPEGIO
    qt = torch.ones((1, 1, 8, 8), dtype=torch.int64)  # From JPEGIO
    model = FFDN()
    pred = model(img, dct, qt, mask)
    print(pred)



"""
ADCD-Net: Adaptive DCT-Guided Content Disentangling Network for Document Tampering Detection

Paper: ICCV 2025
Reference: https://github.com/KAHIMWONG/ADCD-Net

This model combines:
1. Adaptive DCT modulation for handling block misalignment
2. Hierarchical content disentanglement for text-background separation
3. Pristine prototype estimation using background regions
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Dict, Any

from ForensicHub.registry import register_model
from ForensicHub.core.base_model import BaseModel

from .fph import FPH, AddCoord
from .restormer import get_restormer
from .loss import ADCDNetLoss, supcon_parallel


def get_mlp(in_channels, out_channels=None, bias=False):
    """Simple MLP with two 1x1 convolutions."""
    return nn.Sequential(
        nn.Conv2d(in_channels, in_channels // 2, 1, bias=bias),
        nn.GELU(),
        nn.Conv2d(in_channels // 2, out_channels, 1, bias=bias)
    )


@register_model("ADCDNet")
class ADCDNet(BaseModel):
    """
    ADCD-Net for document tampering detection.

    Features:
    - Adaptive DCT modulation with alignment score prediction
    - Hierarchical content disentanglement
    - Pristine prototype estimation from background regions
    - Multi-scale feature extraction
    """

    def __init__(
            self,
            cls_n: int = 2,
            loc_out_dim: int = 96,
            rec_out_dim: int = 4,
            dct_feat_dim: int = 256,
            focal_in_dim: list = [192, 96, 96, 96],
            focal_out_dim: int = 32,
            pp_scale_n: int = 4,
            docres_ckpt_path: str = None,
            ce_weight: float = 3.0,
            rec_weight: float = 1.0,
            focal_weight: float = 0.2,
            norm_weight: float = 0.1,
            use_checkpoint: bool = False,
            **kwargs
    ):
        """
        Args:
            cls_n: Number of classes (2 for binary segmentation)
            loc_out_dim: Output dimension for localization branch
            rec_out_dim: Output dimension for reconstruction (RGB + DCT)
            dct_feat_dim: DCT feature dimension
            focal_in_dim: Input dimensions for focal projection at each scale
            focal_out_dim: Output dimension for focal projection
            pp_scale_n: Number of scales for pristine prototype
            docres_ckpt_path: Path to pre-trained DocRes checkpoint
            ce_weight: Weight for cross-entropy loss
            rec_weight: Weight for reconstruction loss
            focal_weight: Weight for focal contrastive loss
            norm_weight: Weight for normalization loss
            use_checkpoint: Use gradient checkpointing to save memory
        """
        super().__init__()

        self.cls_n = cls_n
        self.docres_ckpt_path = docres_ckpt_path

        # RGB encoder + localization branch (Restormer-based)
        self.restormer_loc = get_restormer(
            model_name='full_model',
            out_channels=loc_out_dim,
            use_checkpoint=use_checkpoint
        )

        # Reconstruction branch (decoder only)
        self.restormer_rec = get_restormer(
            model_name='decoder_only',
            out_channels=rec_out_dim
        )

        # DCT encoder (Feature Pyramid Handler)
        self.dct_encoder = FPH(dct_feat_dim=dct_feat_dim)

        # Alignment score predictor
        self.dct_align_head = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Flatten(),
            nn.Linear(dct_feat_dim, dct_feat_dim // 2, bias=False),
            nn.BatchNorm1d(dct_feat_dim // 2),
            nn.ReLU(inplace=True),
            nn.Linear(dct_feat_dim // 2, 2, bias=False)
        )

        # Coordinate addition for input
        self.add_coord = AddCoord(with_r=False)

        # Focal projection heads for contrastive learning
        self.focal_proj = nn.ModuleList([
            get_mlp(in_channels=focal_in_dim[i], out_channels=focal_out_dim)
            for i in range(len(focal_in_dim))
        ])

        # Pristine prototype estimation
        self.pp_scale_proj = get_mlp(in_channels=pp_scale_n, out_channels=loc_out_dim)
        self.pp_bias_proj = get_mlp(in_channels=pp_scale_n, out_channels=loc_out_dim)

        # Output head
        self.out_head = get_mlp(in_channels=loc_out_dim, out_channels=cls_n)

        # Loss function
        self.loss_fn = ADCDNetLoss(
            ce_weight=ce_weight,
            rec_weight=rec_weight,
            focal_weight=focal_weight,
            norm_weight=norm_weight
        )

        # Load pre-trained weights if available
        if docres_ckpt_path:
            self.load_docres()

    def load_docres(self):
        """Load pre-trained DocRes weights for the encoder."""
        try:
            ckpt = torch.load(self.docres_ckpt_path, map_location='cpu', weights_only=True)
            if 'model_state' in ckpt:
                ckpt = ckpt['model_state']
            elif 'state_dict' in ckpt:
                ckpt = ckpt['state_dict']

            # Remove 'output' layer weights
            for name in list(ckpt.keys()):
                if 'output' in name:
                    ckpt.pop(name)

            # Remove 'module.' prefix if present
            new_ckpt = {}
            for key in ckpt:
                if key.startswith('module.'):
                    new_ckpt[key[7:]] = ckpt[key]
                else:
                    new_ckpt[key] = ckpt[key]

            miss, unexpected = self.restormer_loc.load_state_dict(new_ckpt, strict=False)
            print(f'Loaded DocRes checkpoint from {self.docres_ckpt_path}')
            if miss:
                print(f'  Missing keys: {len(miss)}')
            if unexpected:
                print(f'  Unexpected keys: {len(unexpected)}')
        except Exception as e:
            print(f'Warning: Could not load DocRes checkpoint: {e}')

    def get_pp_map(self, pp_feats, ocr_mask, img_size):
        """
        Compute pristine prototype maps from multi-scale features.

        Args:
            pp_feats: List of features at different scales
            ocr_mask: OCR mask indicating text regions (B, 1, H, W)
            img_size: Target output size

        Returns:
            pp_maps: Pristine prototype maps (B, num_scales, img_size, img_size)
        """
        maps_per_level = []

        # Use OCR mask for background identification
        y = ocr_mask.float()

        for f in pp_feats:
            f = F.normalize(f, p=2, dim=1)  # L2 normalize features
            _, c, h, w = f.shape

            # Resize mask to feature resolution
            bg_mask = (F.interpolate(y, size=(h, w), mode='nearest') == 0).float()

            # Compute background mean feature
            bg_sum = (f * bg_mask).sum(dim=(2, 3))
            bg_count = bg_mask.sum(dim=(2, 3)).clamp_min(1.0)
            bg_mean = F.normalize(bg_sum / bg_count, p=2, dim=1)

            # Compute cosine similarity map
            sim = (f * bg_mean[:, :, None, None]).sum(dim=1, keepdim=True)

            # Upscale to output size
            sim = F.interpolate(sim, size=(img_size, img_size), mode='bilinear', align_corners=False)
            maps_per_level.append(sim)

        return torch.cat(maps_per_level, dim=1)

    def forward(self, image, mask, dct=None, qt=None, ocr_mask=None, **kwargs) -> Dict[str, Any]:
        """
        Forward pass for training.

        Args:
            image: Input RGB image (B, 3, H, W)
            mask: Ground truth tampering mask (B, 1, H, W) or (B, H, W)
            dct: DCT coefficients (B, H, W) with values 0-20
            qt: Quantization table (B, 1, 8, 8) or (B, 8, 8)
            ocr_mask: OCR mask for text regions (B, 1, H, W)

        Returns:
            Dictionary containing:
            - backward_loss: Total loss for backpropagation
            - pred_mask: Predicted tampering mask
            - visual_loss: Individual losses for visualization
            - visual_image: Predicted mask for visualization
        """
        is_train = self.training
        img_size = image.size(2)

        # Ensure mask has correct shape
        if mask.dim() == 3:
            mask = mask.unsqueeze(1)
        mask_for_loss = mask.squeeze(1).long()

        # Handle missing OCR mask
        if ocr_mask is None:
            ocr_mask = torch.zeros_like(mask)

        # Get alignment score & multi-scale DCT features
        if dct is not None and qt is not None:
            as_feat, ms_dct_feats = self.dct_encoder(dct, qt)
            align_logits = self.dct_align_head(as_feat)
            align_score = F.softmax(align_logits, dim=1)[:, -1]
        else:
            ms_dct_feats = None
            align_logits = None
            align_score = None

        # Prepare input with coordinates
        img_with_coord = self.add_coord(image)
        placeholder = torch.zeros_like(ocr_mask).float()
        img_input = torch.cat([img_with_coord, placeholder], dim=1)

        # Get RGB features & localization
        feat, cnt_feats, frg_feats, pp_feats = self.restormer_loc(
            img=img_input,
            ms_dct_feats=ms_dct_feats,
            dct_align_score=align_score
        )

        # Compute pristine prototype maps
        pp_maps = self.get_pp_map(pp_feats, ocr_mask, img_size)
        pp_scale = self.pp_scale_proj(pp_maps)
        pp_bias = self.pp_bias_proj(pp_maps)

        # Apply pristine prototype modulation
        feat = feat * pp_scale + pp_bias

        # Final output
        logits = self.out_head(feat)

        # Training-specific computations
        rec_items = None
        focal_losses = None

        if is_train:
            # Reconstruction with shuffled features
            shuffle_rec_img = self.restormer_rec(cnt_feats, frg_feats, is_shuffle=True)
            if dct is not None:
                norm_dct = (dct.float() / 20.0).unsqueeze(1)
                rec_items = (shuffle_rec_img, norm_dct)

            # Focal contrastive losses
            focal_losses = tuple([
                supcon_parallel(self.focal_proj[i](pp_feats[i]), mask)
                for i in range(len(pp_feats))
            ])

        # Compute loss
        total_loss, loss_dict = self.loss_fn(
            logits=logits,
            mask=mask_for_loss,
            align_logits=align_logits,
            rec_items=rec_items,
            focal_losses=focal_losses
        )

        # Get predicted mask
        pred_mask = F.softmax(logits, dim=1)[:, 1:]  # Take forgery class probability

        output_dict = {
            "backward_loss": total_loss,
            "pred_mask": pred_mask,
            "visual_loss": {
                "seg_loss": loss_dict.get('seg_loss', torch.tensor(0.0)),
                "align_loss": loss_dict.get('align_loss', torch.tensor(0.0)),
                "rec_loss": loss_dict.get('rec_loss', torch.tensor(0.0)),
                "focal_loss": loss_dict.get('focal_loss', torch.tensor(0.0)),
                "combined_loss": total_loss,
            },
            "visual_image": {
                "pred_mask": pred_mask,
            }
        }

        return output_dict

    def get_prediction(self, data_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        Inference without loss computation.

        Args:
            data_dict: Dictionary containing input data

        Returns:
            Dictionary with pred_mask
        """
        self.eval()
        with torch.no_grad():
            image = data_dict['image']
            dct = data_dict.get('dct', None)
            qt = data_dict.get('qt', None)
            ocr_mask = data_dict.get('ocr_mask', None)

            img_size = image.size(2)

            if ocr_mask is None:
                ocr_mask = torch.zeros(image.size(0), 1, image.size(2), image.size(3), device=image.device)

            # DCT features
            if dct is not None and qt is not None:
                as_feat, ms_dct_feats = self.dct_encoder(dct, qt)
                align_logits = self.dct_align_head(as_feat)
                align_score = F.softmax(align_logits, dim=1)[:, -1]
            else:
                ms_dct_feats = None
                align_score = None

            # Prepare input
            img_with_coord = self.add_coord(image)
            placeholder = torch.zeros_like(ocr_mask).float()
            img_input = torch.cat([img_with_coord, placeholder], dim=1)

            # Forward pass
            feat, cnt_feats, frg_feats, pp_feats = self.restormer_loc(
                img=img_input,
                ms_dct_feats=ms_dct_feats,
                dct_align_score=align_score
            )

            # Pristine prototype
            pp_maps = self.get_pp_map(pp_feats, ocr_mask, img_size)
            pp_scale = self.pp_scale_proj(pp_maps)
            pp_bias = self.pp_bias_proj(pp_maps)

            feat = feat * pp_scale + pp_bias
            logits = self.out_head(feat)

            pred_mask = F.softmax(logits, dim=1)[:, 1:]

            return {
                "pred_mask": pred_mask,
            }


if __name__ == "__main__":
    # Test the model
    model = ADCDNet()
    model.train()

    # Create dummy inputs
    batch_size = 2
    img_size = 256
    image = torch.randn(batch_size, 3, img_size, img_size)
    mask = torch.randint(0, 2, (batch_size, 1, img_size, img_size))
    dct = torch.randint(0, 21, (batch_size, img_size, img_size))
    qt = torch.randint(0, 64, (batch_size, 1, 8, 8))
    ocr_mask = torch.randint(0, 2, (batch_size, 1, img_size, img_size))

    # Forward pass
    output = model(image=image, mask=mask, dct=dct, qt=qt, ocr_mask=ocr_mask)

    print("Output keys:", output.keys())
    print("Backward loss:", output['backward_loss'].item())
    print("Pred mask shape:", output['pred_mask'].shape)
    print("Visual losses:", {k: v.item() if torch.is_tensor(v) else v for k, v in output['visual_loss'].items()})



import torch
import torch.nn as nn
import torch.nn.functional as F
import logging
from ForensicHub.registry import register_model
from ForensicHub.core.base_model import BaseModel
import torchvision.models as models
import yaml


@register_model("Capsule_net")
class CapsuleNet(BaseModel):
    def __init__(self, yaml_config_path):
        super(CapsuleNet, self).__init__()

        # 从 YAML 配置文件读取配置
        with open(yaml_config_path, 'r') as file:
            config = yaml.safe_load(file)

        # 配置中的类数量
        self.num_classes = config['num_classes']

        # Initialize CapsuleNet components
        self.vgg_ext = VggExtractor()
        self.fea_ext = FeatureExtractor()
        self.fea_ext.apply(self.weights_init)

        # Routing layer
        self.routing_stats = RoutingLayer(
            num_input_capsules=10,
            num_output_capsules=self.num_classes,
            data_in=8,
            data_out=4,
            num_iterations=2
        )

        # Initialize CapsuleLoss directly
        self.loss_func = CapsuleLoss()  # Use CapsuleLoss directly

    def features(self, image: torch.Tensor) -> torch.tensor:
        """
        Extract features from input image using VGG extractor and feature extractor.

        Args:
            image (torch.Tensor): Input image tensor.

        Returns:
            torch.Tensor: Extracted feature tensor.
        """
        input = self.vgg_ext(image)  # Extract features using VGG extractor
        feature = self.fea_ext(input)  # Further process with feature extractor
        return feature

    def classifier(self, features: torch.tensor) -> torch.tensor:
        """
        Perform classification by applying routing statistics on the features.

        Args:
            features (torch.Tensor): Extracted features from the image.

        Returns:
            torch.Tensor: Class prediction probabilities and mean class.
        """
        z = self.routing_stats(features, random=False, dropout=0.0)
        classes = F.softmax(z, dim=-1)
        class_ = classes.detach()
        class_ = class_.mean(dim=1)
        return classes, class_

    def forward(self, image: torch.Tensor, label: torch.Tensor, **kwargs) -> dict:
        """
        Perform forward propagation.

        Args:
            image (torch.Tensor): Input image tensor.
            label (torch.Tensor): Ground truth label tensor.

        Returns:
            dict: Dictionary containing the backward loss and predictions.
        """
        label = label.long()
        # Extract features from the input image
        features = self.features(image)

        # Get the prediction by classifier
        preds, pred = self.classifier(features)

        # Get the probability of the pred
        prob = torch.softmax(pred, dim=1)[:, 1]

        # Compute the loss using the ground truth labels
        loss = self.loss_func(preds, label)

        # Return prediction and loss values
        return {
            "backward_loss": loss,
            "pred_label": prob,
            "visual_loss": {"combined_loss": loss}
        }

    def weights_init(self, m):
        classname = m.__class__.__name__
        if classname.find('Conv') != -1:
            m.weight.data.normal_(0.0, 0.02)
        elif classname.find('BatchNorm') != -1:
            m.weight.data.normal_(1.0, 0.02)
            m.bias.data.fill_(0)


# VGG input(10,3,256,256)
class VggExtractor(nn.Module):
    def __init__(self, train=False):
        super(VggExtractor, self).__init__()
        self.vgg_1 = self.Vgg(models.vgg19(pretrained=True), 0, 18)
        if train:
            self.vgg_1.train(mode=True)
            self.freeze_gradient()
        else:
            self.vgg_1.eval()

    def Vgg(self, vgg, begin, end):
        features = nn.Sequential(*list(vgg.features.children())[begin:(end + 1)])
        return features

    def freeze_gradient(self, begin=0, end=9):
        for i in range(begin, end + 1):
            self.vgg_1[i].requires_grad = False

    def forward(self, input):
        return self.vgg_1(input)


class FeatureExtractor(nn.Module):
    def __init__(self):
        super(FeatureExtractor, self).__init__()
        self.NO_CAPS = 10
        self.capsules = nn.ModuleList([nn.Sequential(
            nn.Conv2d(256, 64, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.Conv2d(64, 16, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(16),
            nn.ReLU(),
            StatsNet(),
            nn.Conv1d(2, 8, kernel_size=5, stride=2, padding=2),
            nn.BatchNorm1d(8),
            nn.Conv1d(8, 1, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm1d(1),
            View(-1, 8),
        ) for _ in range(self.NO_CAPS)])

    def squash(self, tensor, dim):
        squared_norm = (tensor ** 2).sum(dim=dim, keepdim=True)
        scale = squared_norm / (1 + squared_norm)
        return scale * tensor / (torch.sqrt(squared_norm))

    def forward(self, x):
        outputs = [capsule(x) for capsule in self.capsules]
        output = torch.stack(outputs, dim=-1)
        return self.squash(output, dim=-1)


class StatsNet(nn.Module):
    def __init__(self):
        super(StatsNet, self).__init__()

    def forward(self, x):
        x = x.view(x.data.shape[0], x.data.shape[1], x.data.shape[2] * x.data.shape[3])
        mean = torch.mean(x, 2)
        std = torch.std(x, 2)
        return torch.stack((mean, std), dim=1)


class View(nn.Module):
    def __init__(self, *shape):
        super(View, self).__init__()
        self.shape = shape

    def forward(self, input):
        return input.view(self.shape)


# Capsule right Dynamic routing
class RoutingLayer(nn.Module):
    def __init__(self, num_input_capsules, num_output_capsules, data_in, data_out, num_iterations):
        super(RoutingLayer, self).__init__()

        self.num_iterations = num_iterations
        self.route_weights = nn.Parameter(torch.randn(num_output_capsules, num_input_capsules, data_out, data_in))

    def squash(self, tensor, dim):
        squared_norm = (tensor ** 2).sum(dim=dim, keepdim=True)
        scale = squared_norm / (1 + squared_norm)
        return scale * tensor / (torch.sqrt(squared_norm))

    def forward(self, x, random=False, dropout=0.0):
        x = x.transpose(2, 1)
        if random:
            noise = torch.Tensor(0.01 * torch.randn(*self.route_weights.size())).cuda()
            route_weights = self.route_weights + noise
        else:
            route_weights = self.route_weights
        priors = route_weights[:, None, :, :, :] @ x[None, :, :, :, None]
        priors = priors.transpose(1, 0)
        if dropout > 0.0:
            drop = torch.Tensor(torch.FloatTensor(*priors.size()).bernoulli(1.0 - dropout)).cuda()
            priors = priors * drop
        logits = torch.Tensor(torch.zeros(*priors.size())).to(priors.device)
        for i in range(self.num_iterations):
            probs = F.softmax(logits, dim=2)
            outputs = self.squash((probs * priors).sum(dim=2, keepdim=True), dim=3)
            if i != self.num_iterations - 1:
                delta_logits = priors * outputs
                logits = logits + delta_logits
        outputs = outputs.squeeze()
        if len(outputs.shape) == 3:
            outputs = outputs.transpose(2, 1).contiguous()
        else:
            outputs = outputs.unsqueeze_(dim=0).transpose(2, 1).contiguous()
        return outputs


# Loss Function: CapsuleLoss
class CapsuleLoss(nn.Module):
    def __init__(self):
        super().__init__()
        self.cross_entropy_loss = nn.CrossEntropyLoss()

    def forward(self, inputs, targets):
        """
        Computes the capsule loss.

        Args:
            inputs: A PyTorch tensor of size (batch_size, num_classes) containing the predicted scores.
            targets: A PyTorch tensor of size (batch_size) containing the ground-truth class indices.

        Returns:
            A scalar tensor representing the capsule loss.
        """
        loss_t = self.cross_entropy_loss(inputs[:, 0, :], targets)

        for i in range(inputs.size(1) - 1):
            loss_t = loss_t + self.cross_entropy_loss(inputs[:, i + 1, :], targets)
        return loss_t




import os
from pathlib import Path
import appdirs

class ForHubPath:
    app_name = "ForensicHub"
    app_author = "ForensicHub-authors"

    @classmethod
    def get_data_storage_path(cls):
        storage_path = appdirs.user_data_dir(cls.app_name, cls.app_author)
        if not os.path.exists(storage_path):
            os.makedirs(storage_path)
        return storage_path

    @classmethod
    def get_package_dir(cls):
        return Path(__file__).parent.parent

    @classmethod
    def get_templates_dir(cls):
        return cls.get_package_dir() / 'training_scripts'
    
    @classmethod
    def get_train_test_yaml_dir(cls):
        return cls.get_package_dir() / 'statics'

    

# print("get templates dir: ", ForHubPath.get_templates_dir())


import IMDLBenCo.evaluation
from IMDLBenCo.evaluation import *
from ForensicHub import EVALUATORS
from .AP import ImageAP
from .MCC import ImageMCC
from .TNR import ImageTNR
from .TPR import ImageTPR

# 确保 AbstractEvaluator 被导入，用于后续检查
from IMDLBenCo.evaluation.abstract_class import AbstractEvaluator

# 动态注册所有 AbstractEvaluator 的非抽象子类
for name in dir():
    obj = globals()[name]
    if isinstance(obj, type) and issubclass(obj, AbstractEvaluator):
        # 确保类来自 IMDLBenCo.evaluation 模块
        if obj.__module__.startswith('IMDLBenCo.evaluation'):
            EVALUATORS.register_module(name, force=True, module=obj)

__all__ = IMDLBenCo.evaluation.__all__



import torch

from .guided_diffusion.script_util import (
    create_model_and_diffusion,
)

from ForensicHub.core.base_model import BaseModel
from ForensicHub.registry import register_model
from ForensicHub.common.backbones.resnet import Resnet50

'''
Dire for diffusion-generated image detection
'''


@register_model("Dire")
class Dire(BaseModel):
    def __init__(self,
                 model_path="/mnt/data1/dubo/workspace/ForensicHub/ForensicHub/tasks/aigc/models/dire/imagenet_adm.pth",
                 backbone="resnet50"):
        super().__init__()
        self.backbone = backbone

        self.classifier = None
        if backbone == 'resnet50':
            self.classifier = Resnet50()
        else:
            raise NotImplementedError(f"Backbone {backbone} not supported!")

        use_fp16 = True if self.training else False
        self.model_args = dict(
            attention_resolutions="32,16,8",
            class_cond=False,
            diffusion_steps=1000,
            dropout=0.1,
            image_size=256,
            learn_sigma=True,
            noise_schedule="linear",
            num_channels=256,
            num_head_channels=64,
            num_res_blocks=2,
            resblock_updown=True,
            use_fp16=use_fp16,
            use_scale_shift_norm=True,
            timestep_respacing="ddim20",  # 用于 ddim 重建
            channel_mult="",
            num_heads=4,
            num_heads_upsample=-1,
            use_kl=False,
            predict_xstart=False,
            rescale_timesteps=False,
            rescale_learned_sigmas=False,
            use_checkpoint=False,
            use_new_attention_order=False,
        )
        self.use_ddim = True

        self.model, self.diffusion = create_model_and_diffusion(**self.model_args)
        state_dict = torch.load(model_path, map_location="cpu")['model']
        # self.model.load_state_dict(state_dict)
        if self.model_args["use_fp16"]:
            self.model.convert_to_fp16()
        self.model.eval()

    @torch.no_grad()
    def compute_dire_value(self, image: torch.Tensor) -> torch.Tensor:
        """
        给定图像张量，计算 DiRe 值（图像 - 重构图像）
        """

        latent = self.diffusion.ddim_reverse_sample_loop(
            self.model,
            image.shape,
            noise=image,
            clip_denoised=True,
            model_kwargs={},
            real_step=0
        )

        sample_fn = self.diffusion.ddim_sample_loop if self.use_ddim else self.diffusion.p_sample_loop
        recons = sample_fn(
            self.model,
            (image.size(0), 3, self.model_args["image_size"], self.model_args["image_size"]),
            noise=latent,
            clip_denoised=True,
            model_kwargs={},
            real_step=0
        )

        dire = torch.abs(image - recons)
        return dire

    def forward(self, image: torch.Tensor, label: torch.Tensor, **kwargs):
        label = label.float()

        if kwargs.get('dire') is not None:
            dire = kwargs['dire']
        else:
            dire = self.compute_dire_value(image)

        data_dict = self.classifier(dire, label=label)

        return data_dict



import torch.nn as nn
import torch
import torch.nn.functional as F

from ForensicHub.registry import MODELS, build_from_registry, register_model
from ForensicHub.core.base_model import BaseModel

'''
change data input: [image, mask, label] -> [image, label]
add max pooling detection head and use label for loss only
'''


@register_model("Mask2LabelWrapper")
class Mask2LabelWrapper(BaseModel):
    def __init__(self, name='Resnet50', init_config={}):
        super().__init__()
        self.name = name
        dict = {"name": name, "init_config": init_config}
        self.base_model = build_from_registry(MODELS, dict)
        self.head = nn.AdaptiveMaxPool2d(1)
        self.loss_fn = nn.BCEWithLogitsLoss()

    def forward(self, image, label, *args, **kwargs):
        if not 'mask' in kwargs:
            mask = torch.randint(0, 1, (image.shape[0], 1, image.shape[2], image.shape[3])).long().to(image.device)
        else:
            mask = kwargs['mask']
        if not 'edge_mask' in kwargs:
            edge_mask = torch.randint(0, 1, (image.shape[0], 1, image.shape[2], image.shape[3])).long().to(image.device)
        else:
            edge_mask = kwargs['edge_mask']
        label = label.float()
        if self.name in ['IML_ViT', 'Mesorch', 'Cat_Net']:
            if self.name == 'Cat_Net':
                features = self.base_model.forward_features(image=image, mask=mask, edge_mask=edge_mask, label=label,
                                                            DCT_coef=kwargs['DCT_coef'], qtables=kwargs['qtables'])
            else:
                features = self.base_model.forward_features(image=image, mask=mask, edge_mask=edge_mask, label=label)
            if type(features) == tuple:
                features = features[0]
            pred_label = self.head(features)
            if pred_label.shape[1] != 1:
                pred_label = pred_label[:, 1].view(-1)
            else:
                pred_label = pred_label.view(-1)
            loss = self.loss_fn(pred_label, label.float())
            pred_label = F.sigmoid(pred_label)
        elif self.name in ['DTD', 'FFDN']:
            out_dict = self.base_model(image=image, mask=mask, dct=kwargs['dct'], qt=kwargs['qt'])
            pred_label = self.head(out_dict['pred_mask'][:, 0:1, :, :])
            pred_label = pred_label.view(-1)
            loss = F.binary_cross_entropy(pred_label, label)
        else:
            outputs = self.base_model(image=image, mask=mask, edge_mask=edge_mask, label=label)
            pred_label = outputs['pred_label']
            loss = F.binary_cross_entropy(pred_label, label.float())
        # ----------Output interface--------------------------------------
        output_dict = {
            "backward_loss": loss,
            "pred_label": pred_label,
            "visual_loss": {
                'pred_loss': loss
            }
        }
        return output_dict



import cv2
import torch
from PIL import Image
import numpy as np
import tempfile
import os
import io
from torch.nn import functional as F


def denormalize(image, mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]):
    """denormalize image with mean and std
    """
    image = image.clone().detach().cpu()
    image = image * torch.tensor(std).view(3, 1, 1)
    image = image + torch.tensor(mean).view(3, 1, 1)
    return image



"""
Loss functions for ADCD-Net document tampering detection.

Includes:
- Supervised Contrastive Loss (Focal Loss variant)
- Soft Cross-Entropy Loss with label smoothing
- Lovasz Loss for segmentation
- Combined loss for ADCD-Net training
"""

from typing import Optional
from itertools import filterfalse as ifilterfalse

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pad_sequence


# ========================
# Supervised Contrastive Loss
# ========================

def supcon_parallel(f, y, t=0.1, sample_n=512, min_n=3):
    """
    Supervised contrastive loss computed in parallel across batch.

    Args:
        f: Feature maps (B, C, H, W)
        y: Segmentation masks (B, 1, H, W) or (B, H, W)
        t: Temperature parameter
        sample_n: Number of samples per class
        min_n: Minimum samples required per class

    Returns:
        Contrastive loss scalar
    """
    b, c, h, w = f.shape

    # Reshape y to match f's spatial size
    if y.dim() == 4:
        y = y.squeeze(1)
    y = F.interpolate(y.float().unsqueeze(1), size=(h, w), mode='nearest').long().squeeze(1)

    l = h * w
    f = f.permute(0, 2, 3, 1).reshape(b, l, c)
    y = y.reshape(b, l)

    # Sample features per class
    f_list, y_list, len_list = [], [], []
    for b_idx in range(b):
        bf = f[b_idx]
        by = y[b_idx]
        r_f = bf[by == 0]  # Real/pristine features
        f_f = bf[by == 1]  # Forged features
        r_n, f_n = r_f.size(0), f_f.size(0)

        if r_n < min_n or f_n < min_n:
            continue

        sample_r_f = r_f[torch.randperm(r_f.size(0))[:sample_n]]
        sample_f_f = f_f[torch.randperm(f_f.size(0))[:sample_n]]

        sample_f = torch.cat([sample_r_f, sample_f_f], 0)
        sample_y = torch.cat([torch.zeros(sample_r_f.size(0)), torch.ones(sample_f_f.size(0))], 0)

        f_list.append(sample_f)
        y_list.append(sample_y)
        len_list.append(sample_f.size(0))

    if len(f_list) == 0:
        return torch.tensor([0.0], device=f.device, requires_grad=True).sum()

    pad_f = pad_sequence(f_list, batch_first=True, padding_value=1)
    y = pad_sequence(y_list, batch_first=True, padding_value=-1).to(f.device)
    is_pad = pad_f.sum(-1) == c
    f = F.normalize(pad_f, dim=-1)
    b, l, c = pad_f.shape

    sim = torch.bmm(f, f.permute(0, 2, 1))
    sim = torch.exp(torch.div(sim, t))

    is_valid = ~is_pad
    valid_mask = torch.bmm(is_valid[:, :, None].float(), is_valid[:, None, :].float())

    p_mask = (y[:, None, :] == y[:, :, None]).float()
    eyes = torch.eye(l, dtype=torch.bool, device=f.device).repeat(b, 1, 1)
    reverse_eyes = ~eyes
    p_mask = p_mask * reverse_eyes * valid_mask
    p_num = torch.sum(p_mask, dim=-1)

    n_mask = (y[:, None, :] != y[:, :, None]).float()
    n_mask = n_mask * reverse_eyes * valid_mask

    denominator_p = torch.sum(sim * p_mask, dim=-1, keepdim=True)
    denominator_n = torch.sum(sim * n_mask, dim=1, keepdim=True)
    denominator = denominator_p + denominator_n

    logits = torch.sum(torch.log(sim / (denominator + 1e-8)) * p_mask, dim=-1)
    logits = (-logits / (p_num + 1e-8))
    logits = logits.mean()

    return logits


# ========================
# Soft Cross-Entropy Loss
# ========================

def label_smoothed_nll_loss(
        lprobs: torch.Tensor,
        target: torch.Tensor,
        epsilon: float,
        ignore_index=None,
        reduction="mean",
        dim=-1
) -> torch.Tensor:
    """
    Label smoothed negative log likelihood loss.

    Args:
        lprobs: Log-probabilities of predictions
        target: Target labels
        epsilon: Label smoothing factor
        ignore_index: Index to ignore in loss computation
        reduction: 'mean', 'sum', or 'none'
        dim: Dimension for gathering
    """
    if target.dim() == lprobs.dim() - 1:
        target = target.unsqueeze(dim)

    if ignore_index is not None:
        pad_mask = target.eq(ignore_index)
        target = target.masked_fill(pad_mask, 0)
        nll_loss = -lprobs.gather(dim=dim, index=target)
        smooth_loss = -lprobs.sum(dim=dim, keepdim=True)

        nll_loss = nll_loss.masked_fill(pad_mask, 0.0)
        smooth_loss = smooth_loss.masked_fill(pad_mask, 0.0)
    else:
        nll_loss = -lprobs.gather(dim=dim, index=target)
        smooth_loss = -lprobs.sum(dim=dim, keepdim=True)

        nll_loss = nll_loss.squeeze(dim)
        smooth_loss = smooth_loss.squeeze(dim)

    if reduction == "sum":
        nll_loss = nll_loss.sum()
        smooth_loss = smooth_loss.sum()
    if reduction == "mean":
        nll_loss = nll_loss.mean()
        smooth_loss = smooth_loss.mean()

    eps_i = epsilon / lprobs.size(dim)
    loss = (1.0 - epsilon) * nll_loss + eps_i * smooth_loss
    return loss


class SoftCrossEntropyLoss(nn.Module):
    """Cross-entropy loss with label smoothing."""

    def __init__(
            self,
            reduction: str = "mean",
            smooth_factor: Optional[float] = 0.1,
            ignore_index: Optional[int] = -100,
            dim: int = 1,
    ):
        super().__init__()
        self.smooth_factor = smooth_factor
        self.ignore_index = ignore_index
        self.reduction = reduction
        self.dim = dim

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        log_prob = F.log_softmax(y_pred, dim=self.dim)
        return label_smoothed_nll_loss(
            log_prob,
            y_true,
            epsilon=self.smooth_factor,
            ignore_index=self.ignore_index,
            reduction=self.reduction,
            dim=self.dim,
        )


# ========================
# Lovasz Loss
# ========================

BINARY_MODE = "binary"
MULTICLASS_MODE = "multiclass"
MULTILABEL_MODE = "multilabel"


def _lovasz_grad(gt_sorted):
    """Compute gradient of Lovasz extension."""
    p = len(gt_sorted)
    gts = gt_sorted.sum()
    intersection = gts - gt_sorted.float().cumsum(0)
    union = gts + (1 - gt_sorted).float().cumsum(0)
    jaccard = 1.0 - intersection / union
    if p > 1:
        jaccard[1:p] = jaccard[1:p] - jaccard[0:-1]
    return jaccard


def _lovasz_hinge(logits, labels, per_image=True, ignore=None):
    """Binary Lovasz hinge loss."""
    if per_image:
        loss = mean(
            _lovasz_hinge_flat(*_flatten_binary_scores(log.unsqueeze(0), lab.unsqueeze(0), ignore))
            for log, lab in zip(logits, labels)
        )
    else:
        loss = _lovasz_hinge_flat(*_flatten_binary_scores(logits, labels, ignore))
    return loss


def _lovasz_hinge_flat(logits, labels):
    """Binary Lovasz hinge loss, flat version."""
    if len(labels) == 0:
        return logits.sum() * 0.0
    signs = 2.0 * labels.float() - 1.0
    errors = 1.0 - logits * signs
    errors_sorted, perm = torch.sort(errors, dim=0, descending=True)
    perm = perm.data
    gt_sorted = labels[perm]
    grad = _lovasz_grad(gt_sorted)
    loss = torch.dot(F.relu(errors_sorted), grad)
    return loss


def _flatten_binary_scores(scores, labels, ignore=None):
    """Flatten predictions and labels."""
    scores = scores.view(-1)
    labels = labels.view(-1)
    if ignore is None:
        return scores, labels
    valid = labels != ignore
    vscores = scores[valid]
    vlabels = labels[valid]
    return vscores, vlabels


def _lovasz_softmax(probas, labels, classes="present", per_image=False, ignore=None):
    """Multi-class Lovasz-Softmax loss."""
    if per_image:
        loss = mean(
            _lovasz_softmax_flat(*_flatten_probas(prob.unsqueeze(0), lab.unsqueeze(0), ignore), classes=classes)
            for prob, lab in zip(probas, labels)
        )
    else:
        loss = _lovasz_softmax_flat(*_flatten_probas(probas, labels, ignore), classes=classes)
    return loss


def _lovasz_softmax_flat(probas, labels, classes="present"):
    """Multi-class Lovasz-Softmax loss, flat version."""
    if probas.numel() == 0:
        return probas * 0.0
    C = probas.size(1)
    losses = []
    class_to_sum = list(range(C)) if classes in ["all", "present"] else classes
    for c in class_to_sum:
        fg = (labels == c).type_as(probas)
        if classes == "present" and fg.sum() == 0:
            continue
        if C == 1:
            if len(classes) > 1:
                raise ValueError("Sigmoid output possible only with 1 class")
            class_pred = probas[:, 0]
        else:
            class_pred = probas[:, c]
        errors = (fg - class_pred).abs()
        errors_sorted, perm = torch.sort(errors, 0, descending=True)
        perm = perm.data
        fg_sorted = fg[perm]
        losses.append(torch.dot(errors_sorted, _lovasz_grad(fg_sorted)))
    return mean(losses)


def _flatten_probas(probas, labels, ignore=None):
    """Flatten probabilities and labels for Lovasz loss."""
    if probas.dim() == 3:
        B, H, W = probas.size()
        probas = probas.view(B, 1, H, W)
    C = probas.size(1)
    probas = torch.movedim(probas, 0, -1)
    probas = probas.contiguous().view(-1, C)
    labels = labels.view(-1)
    if ignore is None:
        return probas, labels
    valid = labels != ignore
    vprobas = probas[valid]
    vlabels = labels[valid]
    return vprobas, vlabels


def isnan(x):
    return x != x


def mean(values, ignore_nan=False, empty=0):
    """Compute mean, handling NaN values."""
    values = iter(values)
    if ignore_nan:
        values = ifilterfalse(isnan, values)
    try:
        n = 1
        acc = next(values)
    except StopIteration:
        if empty == "raise":
            raise ValueError("Empty mean")
        return empty
    for n, v in enumerate(values, 2):
        acc += v
    if n == 1:
        return acc
    return acc / n


class LovaszLoss(nn.Module):
    """Lovasz loss for image segmentation."""

    def __init__(
            self,
            mode: str = MULTICLASS_MODE,
            per_image: bool = False,
            ignore_index: Optional[int] = None,
            from_logits: bool = True,
    ):
        assert mode in {BINARY_MODE, MULTILABEL_MODE, MULTICLASS_MODE}
        super().__init__()
        self.mode = mode
        self.ignore_index = ignore_index
        self.per_image = per_image
        self.from_logits = from_logits

    def forward(self, y_pred, y_true):
        if self.mode in {BINARY_MODE, MULTILABEL_MODE}:
            loss = _lovasz_hinge(y_pred, y_true, per_image=self.per_image, ignore=self.ignore_index)
        elif self.mode == MULTICLASS_MODE:
            if self.from_logits:
                y_pred = y_pred.softmax(dim=1)
            loss = _lovasz_softmax(y_pred, y_true, per_image=self.per_image, ignore=self.ignore_index)
        else:
            raise ValueError(f"Wrong mode {self.mode}")
        return loss


# ========================
# Combined ADCD-Net Loss
# ========================

class ADCDNetLoss(nn.Module):
    """
    Combined loss for ADCD-Net training.

    Combines:
    - Segmentation loss (CE + Lovasz)
    - Alignment loss (CE for alignment score prediction)
    - Reconstruction loss (L1)
    - Focal contrastive loss
    """

    def __init__(
            self,
            ce_weight: float = 3.0,
            rec_weight: float = 1.0,
            focal_weight: float = 0.2,
            norm_weight: float = 0.1,
            smooth_factor: float = 0.1,
    ):
        super().__init__()
        self.ce_weight = ce_weight
        self.rec_weight = rec_weight
        self.focal_weight = focal_weight
        self.norm_weight = norm_weight

        self.seg_ce = SoftCrossEntropyLoss(smooth_factor=smooth_factor)
        self.lovasz = LovaszLoss(mode=MULTICLASS_MODE, per_image=True)
        self.align_ce = nn.CrossEntropyLoss()

    def forward(
            self,
            logits: torch.Tensor,
            mask: torch.Tensor,
            align_logits: torch.Tensor = None,
            rec_items: tuple = None,
            focal_losses: tuple = None,
    ):
        """
        Compute combined loss.

        Args:
            logits: Segmentation logits (B, 2, H, W)
            mask: Ground truth mask (B, H, W) or (B, 1, H, W)
            align_logits: Alignment prediction logits (B, 2)
            rec_items: Tuple of (reconstructed, target) for reconstruction loss
            focal_losses: Tuple of focal contrastive losses at multiple scales

        Returns:
            total_loss: Combined loss
            loss_dict: Dictionary of individual losses for logging
        """
        # Ensure mask has correct shape
        if mask.dim() == 4:
            mask = mask.squeeze(1)
        mask = mask.long()

        # Resize logits to match mask if needed
        if logits.shape[-2:] != mask.shape[-2:]:
            logits = F.interpolate(logits, size=mask.shape[-2:], mode='bilinear', align_corners=False)

        # Segmentation loss
        seg_ce_loss = self.seg_ce(logits, mask)
        seg_lovasz_loss = self.lovasz(logits, mask)
        seg_loss = seg_ce_loss + seg_lovasz_loss

        total_loss = self.ce_weight * seg_loss
        loss_dict = {
            'seg_ce_loss': seg_ce_loss,
            'seg_lovasz_loss': seg_lovasz_loss,
            'seg_loss': seg_loss,
        }

        # Alignment loss
        if align_logits is not None:
            # Create alignment targets: 1 if image has tampering, 0 otherwise
            align_target = (mask.sum(dim=(1, 2)) > 0).long()
            align_loss = self.align_ce(align_logits, align_target)
            total_loss = total_loss + align_loss
            loss_dict['align_loss'] = align_loss

        # Reconstruction loss
        if rec_items is not None:
            rec_pred, rec_target = rec_items
            # Split into RGB and DCT components
            rec_rgb = rec_pred[:, :3]
            rec_dct = rec_pred[:, 3:]

            # L1 loss for reconstruction
            if rec_target.dim() == 4 and rec_target.shape[1] == 1:
                # DCT-only target
                rec_loss = F.l1_loss(rec_dct, rec_target)
            else:
                # Full reconstruction
                rec_loss = F.l1_loss(rec_pred, rec_target)

            total_loss = total_loss + self.rec_weight * rec_loss
            loss_dict['rec_loss'] = rec_loss

        # Focal contrastive loss
        if focal_losses is not None:
            focal_loss = sum(focal_losses) / len(focal_losses)
            total_loss = total_loss + self.focal_weight * focal_loss
            loss_dict['focal_loss'] = focal_loss

        loss_dict['total_loss'] = total_loss

        return total_loss, loss_dict



from .version import __version__, version_info
from .registry import Registry, MODELS, DATASETS, POSTFUNCS, TRANSFORMS, EVALUATORS
from .common.backbones import *
from .common.evaluation import *

from .tasks import *
from .core import *
# make sure wrapper to be registered after models
from .common.wrapper import *

__all__ = ['__version__', 'version_info', 'MODELS', "DATASETS", "POSTFUNCS", "TRANSFORMS", "EVALUATORS"]



def get_all_lazy_maps():
    # 在函数内部导入，避免循环
    from ForensicHub.tasks.aigc.models import _lazy_model_map as _aigc_model_map
    from ForensicHub.tasks.aigc.models import _lazy_postfunc_map as _aigc_postfunc_map
    from ForensicHub.tasks.document.models import _lazy_model_map as _document_model_map
    from ForensicHub.tasks.document.models import _lazy_postfunc_map as _document_postfunc_map
    from ForensicHub.tasks.deepfake import _lazy_model_map as _deepfake_model_map
    from ForensicHub.tasks.deepfake import _lazy_postfunc_map as _deepfake_postfunc_map

    model_map = {}
    model_map.update(_aigc_model_map)
    model_map.update(_document_model_map)
    model_map.update(_deepfake_model_map)

    postfunc_map = {}
    postfunc_map.update(_aigc_postfunc_map)
    postfunc_map.update(_document_postfunc_map)
    postfunc_map.update(_deepfake_postfunc_map)

    return model_map, postfunc_map


def _wrap_deepfake_if_needed(cls, name):
    """如果是 deepfake 模型，则包一层 Wrapper"""
    from ForensicHub.tasks.deepfake import _lazy_model_map as _deepfake_model_map
    from ForensicHub.tasks.deepfake import _lazy_postfunc_map as _deepfake_postfunc_map
    from ForensicHub.tasks.deepfake.wrapper.wrappers import Deepfake2ForensicWrapper
    from DeepfakeBench.training import detectors

    flag = False

    if name in _deepfake_model_map:
        flag = True
        cls = getattr(detectors, name)

        class WrappedModel(Deepfake2ForensicWrapper):
            def __init__(self, *args, **kwargs):
                super().__init__(cls, *args, **kwargs)

        WrappedModel.__name__ = f"Wrapped{name}"
        WrappedModel.__qualname__ = WrappedModel.__name__
        return WrappedModel, flag

    return cls, flag



import json
import os
from typing import Dict, Any, List, Optional
import torch
from torch.utils.data import Dataset
from ForensicHub.core.base_dataset import BaseDataset
from ForensicHub.registry import register_dataset
from PIL import Image
import numpy as np


@register_dataset("DireDataset")
class DireDataset(BaseDataset):
    """Dataset for Dire model."""

    def __init__(self,
                 path: str,
                 image_size: int = 224,
                 gen_mask: bool = False,
                 **kwargs):
        self.image_size = image_size
        self.gen_mask = gen_mask
        super().__init__(path=path, **kwargs)

    def _init_dataset_path(self) -> None:
        """Read JSON file and parse image paths and labels."""
        if not os.path.exists(self.path):
            raise FileNotFoundError(f"JSON file not found at {self.path}")

        with open(self.path, 'r') as f:
            data = json.load(f)

        self.samples = data  # Expecting a list of {"path": ..., "label": ...}
        self.entry_path = self.path  # For __str__()

    def __len__(self) -> int:
        return len(self.samples)

    def __getitem__(self, idx: int) -> Dict[str, torch.Tensor]:
        sample = self.samples[idx]
        image_path = sample["path"]
        label = sample["label"]

        # Resolve full path if needed
        if not os.path.isabs(image_path):
            image_path = os.path.join(os.path.dirname(self.path), image_path)

        # Load image
        image = Image.open(image_path).convert("RGB")
        image = image.resize((self.image_size, self.image_size))
        image = np.array(image)

        # Apply transforms
        if self.common_transform:
            image = self.common_transform(image=image)['image']
        if self.post_transform:
            image = self.post_transform(image=image)['image']

        output = {
            "image": image,
            "dire": image,
            "label": torch.tensor(label, dtype=torch.float)
        }

        if self.gen_mask:
            mask = torch.full((image.shape[1], image.shape[2]), fill_value=label, dtype=torch.float)
            output["mask"] = mask

        # Apply post-processing functions
        if self.post_funcs:
            self.post_funcs(output)

        return output



import torch
import torch.nn as nn
from ..utils.srm_filter_kernel import *
import torch.nn.functional as F


class Block_(nn.Module):
    def __init__(self, in_channels, out_channels, depth):
        super(Block_, self).__init__()
        self.process = nn.ModuleList([nn.Conv2d(in_channels, out_channels, kernel_size=(3, 3), stride=(1, 1), padding=1, bias=False)])
        for i in range(depth-1):
            self.process.append(nn.Conv2d(out_channels, out_channels, kernel_size=(3, 3), stride=(1, 1), padding=1, bias=False))

    def forward(self, x):
        out = x
        for conv in self.process:
            out = conv(x)
        return out

class Pre4(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 3, 1, 1, 0)
        self.blk2 = Block_(6, 6, depth=1)
        self.blk3 = Block_(6, 6, depth=1)

    def forward(self, x):
        f1 = self.conv1(x)
        f1 = torch.cat((f1, x), dim=1)
        f2 = self.blk2(f1)
        f = f1-f2
        f3 = self.blk3(f)
        f = torch.cat((f, f3), dim=1)
        return f

class Block2(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(Block2, self).__init__()
        self.process = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=(3, 3), stride=(1, 1), padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, kernel_size=(3, 3), stride=(1, 1), padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(3,2,1)
        )

    def forward(self, x):
        out = self.process(x)
        return out

class plain(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(plain, self).__init__()
        self.blk = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=(3, 3), stride=(1, 1), padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(3,2,1)
        )
    def forward(self, x):
        x = self.blk(x)
        return x


class Pre2(nn.Module):
    def __init__(self):
        super(Pre2, self).__init__()
        self.pre = Pre4()
        self.layers = nn.Sequential(
            Block2(12, 64),
            plain(64, 128),
            plain(128, 256),
        )


    def forward(self, inp):
        b, _, h, w = inp.size()
        x = self.pre(inp)
        x = self.layers(x)
        return x



import torch
import jpegio
import numpy as np
import torch.nn as nn
import torch.optim as optim
import torch.distributed as dist
from torch.autograd import Variable
from .fph import FPH
from .backbone_swin import SwinTransformerV2
from .backbone_convnext import ConvNeXt
import torch.nn.functional as F
from timm.models.layers import trunc_normal_, DropPath
from functools import partial
from typing import Optional, Union, List, Dict, Any

from ForensicHub.registry import register_model
from ForensicHub.core.base_model import BaseModel

'''
Towards Robust Tampered Text Detection in Document Image: New dataset and New Solution
https://openaccess.thecvf.com/content/CVPR2023/papers/Qu_Towards_Robust_Tampered_Text_Detection_in_Document_Image_New_Dataset_CVPR_2023_paper.pdf
'''


class SoftDiceLossV1(nn.Module):
    '''
    soft-dice loss, useful in binary segmentation
    '''

    def __init__(self,
                 p=1,
                 smooth=1):
        super(SoftDiceLossV1, self).__init__()
        self.p = p
        self.smooth = smooth

    def forward(self, logits, labels):
        '''
        inputs:
            logits: tensor of shape (N, H, W, ...)
            label: tensor of shape(N, H, W, ...)
        output:
            loss: tensor of shape(1, )
        '''
        probs = F.softmax(logits, 1)[:, 1]
        numer = (probs * labels).sum()
        denor = (probs.pow(self.p) + labels.pow(self.p)).sum()
        loss = 1. - (2 * numer + self.smooth) / (denor + self.smooth)
        return loss


class LayerNorm(nn.Module):
    def __init__(self, normalized_shape, eps=1e-6, data_format="channels_last"):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.eps = eps
        self.data_format = data_format
        if self.data_format not in ["channels_last", "channels_first"]:
            raise NotImplementedError
        self.normalized_shape = (normalized_shape,)

    def forward(self, x):
        if self.data_format == "channels_last":
            return F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        elif self.data_format == "channels_first":
            u = x.mean(1, keepdim=True)
            s = (x - u).pow(2).mean(1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.eps)
            x = self.weight[:, None, None] * x + self.bias[:, None, None]
            return x


def get_norm_layer(
        use_norm: Union[bool, str, Dict[str, Any]], out_channels: int
) -> nn.Module:
    supported_norms = ("inplace", "batchnorm", "identity", "layernorm", "instancenorm")

    # Step 1. Convert tot dict representation

    ## Check boolean
    if use_norm is True:
        norm_params = {"type": "batchnorm"}
    elif use_norm is False:
        norm_params = {"type": "identity"}

    ## Check string
    elif isinstance(use_norm, str):
        norm_str = use_norm.lower()
        if norm_str == "inplace":
            norm_params = {
                "type": "inplace",
                "activation": "leaky_relu",
                "activation_param": 0.0,
            }
        elif norm_str in supported_norms:
            norm_params = {"type": norm_str}
        else:
            raise ValueError(
                f"Unrecognized normalization type string provided: {use_norm}. Should be in "
                f"{supported_norms}"
            )

    ## Check dict
    elif isinstance(use_norm, dict):
        norm_params = use_norm

    else:
        raise ValueError(
            f"Invalid type for use_norm should either be a bool (batchnorm/identity), "
            f"a string in {supported_norms}, or a dict like {{'type': 'batchnorm', **kwargs}}"
        )

    # Step 2. Check if the dict is valid
    if "type" not in norm_params:
        raise ValueError(
            f"Malformed dictionary given in use_norm: {use_norm}. Should contain key 'type'."
        )
    if norm_params["type"] not in supported_norms:
        raise ValueError(
            f"Unrecognized normalization type string provided: {use_norm}. Should be in {supported_norms}"
        )
    if norm_params["type"] == "inplace" and InPlaceABN is None:
        raise RuntimeError(
            "In order to use `use_norm='inplace'` the inplace_abn package must be installed. Use:\n"
            "  $ pip install -U wheel setuptools\n"
            "  $ pip install inplace_abn --no-build-isolation\n"
            "Also see: https://github.com/mapillary/inplace_abn"
        )

    # Step 3. Initialize the norm layer
    norm_type = norm_params["type"]
    norm_kwargs = {k: v for k, v in norm_params.items() if k != "type"}

    if norm_type == "inplace":
        norm = InPlaceABN(out_channels, **norm_kwargs)
    elif norm_type == "batchnorm":
        norm = nn.BatchNorm2d(out_channels, **norm_kwargs)
    elif norm_type == "identity":
        norm = nn.Identity()
    elif norm_type == "layernorm":
        norm = nn.LayerNorm(out_channels, **norm_kwargs)
    elif norm_type == "instancenorm":
        norm = nn.InstanceNorm2d(out_channels, **norm_kwargs)
    else:
        raise ValueError(f"Unrecognized normalization type: {norm_type}")

    return norm


class Conv2dReLU(nn.Sequential):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: int,
            padding: int = 0,
            stride: int = 1,
            use_batchnorm: bool = True,
    ):
        if use_batchnorm:
            norm = get_norm_layer(use_batchnorm, out_channels)

        is_identity = isinstance(norm, nn.Identity)
        conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            bias=is_identity,
        )

        activation = nn.ReLU(inplace=True)

        super(Conv2dReLU, self).__init__(conv, norm, activation)


class SCSEModule(nn.Module):
    def __init__(self, in_channels, reduction=16):
        super().__init__()
        self.cSE = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_channels, in_channels // reduction, 1),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels // reduction, in_channels, 1),
            nn.Sigmoid(),
        )
        self.sSE = nn.Sequential(nn.Conv2d(in_channels, 1, 1), nn.Sigmoid())

    def forward(self, x):
        return x * self.cSE(x) + x * self.sSE(x)


class ConvBlock(nn.Module):
    def __init__(self, dim, drop_path=0., layer_scale_init_value=1e-6):
        super().__init__()
        self.dwconv = nn.Conv2d(dim, dim, kernel_size=7, padding=3, groups=dim)
        self.norm = LayerNorm(dim, eps=1e-6)
        self.pwconv1 = nn.Linear(dim, 4 * dim)
        self.act = nn.GELU()
        self.pwconv2 = nn.Linear(4 * dim, dim)
        self.gamma = nn.Parameter(layer_scale_init_value * torch.ones((dim)),
                                  requires_grad=True) if layer_scale_init_value > 0 else None
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

    def forward(self, x):
        ipt = x
        x = self.dwconv(x)
        x = x.permute(0, 2, 3, 1)
        x = self.norm(x)
        x = self.pwconv1(x)
        x = self.act(x)
        x = self.pwconv2(x)
        if self.gamma is not None:
            x = self.gamma * x
        x = x.permute(0, 3, 1, 2)
        x = ipt + self.drop_path(x)
        return x


class AddCoords(nn.Module):
    def __init__(self, with_r=True):
        super().__init__()
        self.with_r = with_r

    def forward(self, input_tensor):
        batch_size, _, x_dim, y_dim = input_tensor.size()
        xx_c, yy_c = torch.meshgrid(torch.arange(x_dim, dtype=input_tensor.dtype),
                                    torch.arange(y_dim, dtype=input_tensor.dtype))
        xx_c = xx_c.to(input_tensor.device) / (x_dim - 1) * 2 - 1
        yy_c = yy_c.to(input_tensor.device) / (y_dim - 1) * 2 - 1
        xx_c = xx_c.expand(batch_size, 1, x_dim, y_dim)
        yy_c = yy_c.expand(batch_size, 1, x_dim, y_dim)
        ret = torch.cat((input_tensor, xx_c, yy_c), dim=1)
        if self.with_r:
            rr = torch.sqrt(torch.pow(xx_c - 0.5, 2) + torch.pow(yy_c - 0.5, 2))
            ret = torch.cat([ret, rr], dim=1)
        return ret


class VPH(nn.Module):
    def __init__(self, dims=[96, 192], drop_path_rate=0.4, layer_scale_init_value=1e-6):
        super().__init__()
        dp_rates = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]
        self.downsample_layers = nn.ModuleList([nn.Sequential(nn.Conv2d(6, dims[0], kernel_size=4, stride=4),
                                                              LayerNorm(dims[0], eps=1e-6,
                                                                        data_format="channels_first")), nn.Sequential(
            LayerNorm(dims[1], eps=1e-6, data_format="channels_first"),
            nn.Conv2d(dims[1], dims[2], kernel_size=2, stride=2))])
        self.stages = nn.ModuleList([nn.Sequential(
            *[ConvBlock(dim=dims[0], drop_path=dp_rates[j], layer_scale_init_value=layer_scale_init_value) for j in
              range(3)]), nn.Sequential(
            *[ConvBlock(dim=dims[1], drop_path=dp_rates[3 + j], layer_scale_init_value=layer_scale_init_value) for j in
              range(3)])])
        self.apply(self._init_weights)

    def initnorm(self):
        norm_layer = partial(LayerNorm, eps=1e-6, data_format="channels_first")
        for i_layer in range(4):
            layer = norm_layer(self.dims[i_layer])
            layer_name = f'norm{i_layer}'
            self.add_module(layer_name, layer)

    def _init_weights(self, m):
        if isinstance(m, (nn.Conv2d, nn.Linear)):
            trunc_normal_(m.weight, std=.02)
            nn.init.constant_(m.bias, 0)

    def init_weights(self, pretrained=None):
        def _init_weights(m):
            if isinstance(m, nn.Linear):
                trunc_normal_(m.weight, std=.02)
                if isinstance(m, nn.Linear) and m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

        self.apply(_init_weights)

    def forward(self, x):
        outs = []
        x = self.stages[0](self.downsample_layers[0](x))
        outs = [self.norm0(x)]
        x = self.stages[1](self.downsample_layers[1](x))
        outs.append(self.norm1(x))
        return outs


class DecoderBlock(nn.Module):
    def __init__(self, cin, cadd, cout, ):
        super().__init__()
        self.cin = (cin + cadd)
        self.cout = cout
        self.conv1 = Conv2dReLU(self.cin, self.cout, kernel_size=3, padding=1, use_batchnorm=True)
        self.conv2 = Conv2dReLU(self.cout, self.cout, kernel_size=3, padding=1, use_batchnorm=True)

    def forward(self, x1, x2=None):
        x1 = F.interpolate(x1, scale_factor=2.0, mode="nearest")
        if x2 is not None:
            x1 = torch.cat([x1, x2], dim=1)
        x1 = self.conv1(x1[:, :self.cin])
        x1 = self.conv2(x1)
        return x1


class ConvBNReLU(nn.Module):
    def __init__(self, in_c, out_c, ks, stride=1, norm=True, res=False):
        super(ConvBNReLU, self).__init__()
        if norm:
            self.conv = nn.Sequential(
                nn.Conv2d(in_c, out_c, kernel_size=ks, padding=ks // 2, stride=stride, bias=False),
                nn.BatchNorm2d(out_c), nn.ReLU(True))
        else:
            self.conv = nn.Conv2d(in_c, out_c, kernel_size=ks, padding=ks // 2, stride=stride, bias=False)
        self.res = res

    def forward(self, x):
        if self.res:
            return (x + self.conv(x))
        else:
            return self.conv(x)


class FUSE1(nn.Module):
    def __init__(self, in_channels_list=(96, 192, 384, 768)):
        super(FUSE1, self).__init__()
        self.c31 = ConvBNReLU(in_channels_list[2], in_channels_list[2], 1)
        self.c32 = ConvBNReLU(in_channels_list[3], in_channels_list[2], 1)
        self.c33 = ConvBNReLU(in_channels_list[2], in_channels_list[2], 3)

        self.c21 = ConvBNReLU(in_channels_list[1], in_channels_list[1], 1)
        self.c22 = ConvBNReLU(in_channels_list[2], in_channels_list[1], 1)
        self.c23 = ConvBNReLU(in_channels_list[1], in_channels_list[1], 3)

        self.c11 = ConvBNReLU(in_channels_list[0], in_channels_list[0], 1)
        self.c12 = ConvBNReLU(in_channels_list[1], in_channels_list[0], 1)
        self.c13 = ConvBNReLU(in_channels_list[0], in_channels_list[0], 3)

    def forward(self, x):
        x, x1, x2, x3 = x
        h, w = x2.shape[-2:]
        x2 = self.c33(F.interpolate(self.c32(x3), size=(h, w)) + self.c31(x2))
        h, w = x1.shape[-2:]
        x1 = self.c23(F.interpolate(self.c22(x2), size=(h, w)) + self.c21(x1))
        h, w = x.shape[-2:]
        x = self.c13(F.interpolate(self.c12(x1), size=(h, w)) + self.c11(x))
        return x, x1, x2, x3


class FUSE2(nn.Module):
    def __init__(self, in_channels_list=(96, 192, 384)):
        super(FUSE2, self).__init__()

        self.c21 = ConvBNReLU(in_channels_list[1], in_channels_list[1], 1)
        self.c22 = ConvBNReLU(in_channels_list[2], in_channels_list[1], 1)
        self.c23 = ConvBNReLU(in_channels_list[1], in_channels_list[1], 3)

        self.c11 = ConvBNReLU(in_channels_list[0], in_channels_list[0], 1)
        self.c12 = ConvBNReLU(in_channels_list[1], in_channels_list[0], 1)
        self.c13 = ConvBNReLU(in_channels_list[0], in_channels_list[0], 3)

    def forward(self, x):
        x, x1, x2 = x
        h, w = x1.shape[-2:]
        x1 = self.c23(F.interpolate(self.c22(x2), size=(h, w), mode='bilinear', align_corners=True) + self.c21(x1))
        h, w = x.shape[-2:]
        x = self.c13(F.interpolate(self.c12(x1), size=(h, w), mode='bilinear', align_corners=True) + self.c11(x))
        return x, x1, x2


class FUSE3(nn.Module):
    def __init__(self, in_channels_list=(96, 192)):
        super(FUSE3, self).__init__()

        self.c11 = ConvBNReLU(in_channels_list[0], in_channels_list[0], 1)
        self.c12 = ConvBNReLU(in_channels_list[1], in_channels_list[0], 1)
        self.c13 = ConvBNReLU(in_channels_list[0], in_channels_list[0], 3)

    def forward(self, x):
        x, x1 = x
        h, w = x.shape[-2:]
        x = self.c13(F.interpolate(self.c12(x1), size=(h, w), mode='bilinear', align_corners=True) + self.c11(x))
        return x, x1


class MID(nn.Module):
    def __init__(self, encoder_channels, decoder_channels):
        super().__init__()
        encoder_channels = encoder_channels[1:][::-1]
        self.in_channels = [encoder_channels[0]] + list(decoder_channels[:-1])
        self.add_channels = list(encoder_channels[1:]) + [96]
        self.out_channels = decoder_channels
        self.fuse1 = FUSE1()
        self.fuse2 = FUSE2()
        self.fuse3 = FUSE3()
        decoder_convs = {}
        for layer_idx in range(len(self.in_channels) - 1):
            for depth_idx in range(layer_idx + 1):
                if depth_idx == 0:
                    in_ch = self.in_channels[layer_idx]
                    skip_ch = self.add_channels[layer_idx] * (layer_idx + 1)
                    out_ch = self.out_channels[layer_idx]
                else:
                    out_ch = self.add_channels[layer_idx]
                    skip_ch = self.add_channels[layer_idx] * (layer_idx + 1 - depth_idx)
                    in_ch = self.add_channels[layer_idx - 1]
                decoder_convs[f"x_{depth_idx}_{layer_idx}"] = DecoderBlock(in_ch, skip_ch, out_ch)
        decoder_convs[f"x_{0}_{len(self.in_channels) - 1}"] = DecoderBlock(self.in_channels[-1], 0,
                                                                           self.out_channels[-1])
        self.decoder_convs = nn.ModuleDict(decoder_convs)

    def forward(self, *features):
        decoder_features = {}
        features = self.fuse1(features)[::-1]
        decoder_features["x_0_0"] = self.decoder_convs["x_0_0"](features[0], features[1])
        decoder_features["x_1_1"] = self.decoder_convs["x_1_1"](features[1], features[2])
        decoder_features["x_2_2"] = self.decoder_convs["x_2_2"](features[2], features[3])
        decoder_features["x_2_2"], decoder_features["x_1_1"], decoder_features["x_0_0"] = self.fuse2(
            (decoder_features["x_2_2"], decoder_features["x_1_1"], decoder_features["x_0_0"]))
        decoder_features["x_0_1"] = self.decoder_convs["x_0_1"](decoder_features["x_0_0"],
                                                                torch.cat((decoder_features["x_1_1"], features[2]), 1))
        decoder_features["x_1_2"] = self.decoder_convs["x_1_2"](decoder_features["x_1_1"],
                                                                torch.cat((decoder_features["x_2_2"], features[3]), 1))
        decoder_features["x_1_2"], decoder_features["x_0_1"] = self.fuse3(
            (decoder_features["x_1_2"], decoder_features["x_0_1"]))
        decoder_features["x_0_2"] = self.decoder_convs["x_0_2"](decoder_features["x_0_1"], torch.cat(
            (decoder_features["x_1_2"], decoder_features["x_2_2"], features[3]), 1))
        return self.decoder_convs["x_0_3"](
            torch.cat((decoder_features["x_0_2"], decoder_features["x_1_2"], decoder_features["x_2_2"]), 1))


@register_model("DTD")
class DTD(BaseModel):
    def __init__(self, decoder_channels=(384, 192, 96, 64), classes=2, use_dice_loss=False,
                 convnext_path='/mnt/data1/dubo/workspace/ForensicHub/ForensicHub/tasks/document/models/dtd/convnext_small.pth',
                 swin_path='/mnt/data1/dubo/workspace/ForensicHub/ForensicHub/tasks/document/models/dtd/swintransformerv2_small.pth'):
        super().__init__()
        self.convnext_path = convnext_path
        self.swin_path = swin_path
        self.vph = ConvNeXt(in_chans=6)  # Backbone1, using stage 0, 1
        self.swin = SwinTransformerV2()  # Backbone2, using stages 1, 2, 3
        self.fph = FPH()
        self.decoder = MID(encoder_channels=(96, 192, 384, 768), decoder_channels=decoder_channels)
        self.head = nn.Sequential(nn.Conv2d(decoder_channels[-1], 256, 3, 1, 1),
                                  nn.UpsamplingBilinear2d(scale_factor=2.0), nn.ReLU(), nn.Conv2d(256, 2, 1, 1, 0))
        self.addcoords = AddCoords()
        self.FU = nn.Sequential(SCSEModule(448), nn.Conv2d(448, 192, 3, 1, 1), nn.BatchNorm2d(192), nn.ReLU(True))
        self.init_vph()
        self.init_swin()
        self.initialize_decoder()
        self.use_dice_loss = use_dice_loss
        if use_dice_loss:
            self.dice = SoftDiceLossV1()

    def init_vph(self, ):
        weights = torch.load(self.convnext_path)['state_dict']
        weights['backbone.downsample_layers.0.0.weight'] = torch.cat(
            (weights['backbone.downsample_layers.0.0.weight'], weights['backbone.downsample_layers.0.0.weight']), 1)
        dels = [k for k in weights.keys() if not k.startswith('backbone.')]
        for k in dels:
            del weights[k]
        new_weights = {}
        for k, v in weights.items():
            new_weights['.'.join(k.split('.')[1:])] = v
        self.vph.load_state_dict(new_weights)
        del self.vph.downsample_layers[3]
        del self.vph.stages[3]
        del self.vph.downsample_layers[2]
        del self.vph.stages[2]

    def init_swin(self, ):
        weights = torch.load(self.swin_path)['model']
        self.swin.load_state_dict(weights)
        del self.swin.norm
        del self.swin.patch_embed
        del self.swin.head
        del self.swin.layers[0]

    def initialize_decoder(self, ):
        for m in self.decoder.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_uniform_(m.weight, mode="fan_in", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

            elif isinstance(
                    m, (nn.BatchNorm2d, nn.LayerNorm, nn.GroupNorm, nn.InstanceNorm2d)
            ):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

            elif isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

        for m in self.head.modules():
            if isinstance(m, (nn.Linear, nn.Conv2d)):
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def cal_seg_loss(self, pred, gt):
        h, w = gt.shape[-2:]
        pred = F.interpolate(pred, size=(h, w), mode='bilinear')
        ce_loss = F.cross_entropy(pred, gt) + self.dice(pred, gt) if self.use_dice_loss else F.cross_entropy(pred, gt)
        return ce_loss, pred

    def forward(self, image, dct, qt, mask, **kwargs):
        DCT_coef = dct
        qtables = qt
        x = image
        mask = mask.squeeze(1).long()  # [B,1,H,W] -> [B,H,W]
        DCT_coef = DCT_coef.squeeze(1).long()  # [B,1,H,W] -> [B,H,W]
        if len(qtables.shape) == 3:
            qtables = qtables.unsqueeze(1)
        features = self.vph.forward_features(self.addcoords(x), end_index=2)
        features[1] = self.FU(torch.cat((features[1], self.fph(DCT_coef, qtables)), 1))
        rst = self.swin.layers[0](features[1].flatten(2).transpose(1, 2).contiguous())
        N, L, C = rst.shape
        H = W = int(L ** (1 / 2))
        features.append(self.vph.norm2(rst.transpose(1, 2).contiguous().view(N, C, H, W)))
        features.append(self.vph.norm3(
            self.swin.layers[2](self.swin.layers[1](rst)).transpose(1, 2).contiguous().view(N, C * 2, H // 2, W // 2)))
        decoder_output = self.decoder(*features)
        output = self.head(decoder_output)
        seg_loss, output = self.cal_seg_loss(output, mask)
        output = F.softmax(output, dim=1)
        output = output[:, 1:]
        output_dict = {
            "backward_loss": seg_loss,
            "pred_mask": output,
            "visual_loss": {
                "seg_loss": seg_loss,
                "combined_loss": seg_loss
            },
            "visual_image": {
                "pred_mask": output,
            }
        }
        return output_dict


if __name__ == "__main__":
    img = torch.ones((1, 3, 512, 512))
    mask = torch.ones((1, 1, 512, 512), dtype=torch.int64)
    dct = torch.ones((1, 1, 512, 512), dtype=torch.int64)  # From JPEGIO
    qt = torch.ones((1, 1, 8, 8), dtype=torch.int64)  # From JPEGIO
    model = DTD()
    pred = model(img, dct, qt, mask)
    print(pred)



import json
import os
import random
import torch
import numpy as np
from PIL import Image
from ForensicHub.core.base_dataset import BaseDataset
from ForensicHub.registry import register_dataset


@register_dataset("DeepfakeCrossDataset")
class DeepfakeCrossDataset(BaseDataset):
    def __init__(self, path, config_file, compression='c23', image_size=256, split_mode='train', **kwargs):
        """
        Args:
            path (str): 根目录（包含图像文件的根路径，如 FaceForensics++/original_sequences/...）
            config_file (str): 指向 JSON 配置文件路径
            compression (str): 压缩类型，如 'c23' 或 'c40'
            image_size (int): 图像大小
        """
        self.path = path
        self.config_file = config_file
        self.compression = compression
        self.image_size = image_size
        self.split_mode = split_mode.lower()
        super().__init__(path=path, **kwargs)

    def _init_dataset_path(self):
        """从配置文件中读取图像路径和标签，支持 train/test 分支，压缩率字段可选。"""
        with open(self.config_file, 'r') as f:
            dataset_info = json.load(f)

        samples = []

        for dataset_name, label_groups in dataset_info.items():
            for label_str, splits in label_groups.items():
                label = 0 if "real" in label_str.lower() else 1  # 标签0=真实，1=伪造

                if self.split_mode not in splits:
                    continue
                # 如果有压缩率字段就使用，没有就直接取该分支
                if isinstance(splits[self.split_mode], dict) and self.compression in splits[self.split_mode]:
                    video_group = splits[self.split_mode][self.compression]
                elif isinstance(splits[self.split_mode], dict):
                    video_group = splits[self.split_mode]
                else:
                    continue  # 无法匹配则跳过

                for video_id, video_info in video_group.items():
                    frame_paths = video_info["frames"]
                    for rel_path in frame_paths:
                        full_path = os.path.join(self.path, rel_path)
                        samples.append({
                            "path": full_path,
                            "mask": None,
                            "label": label
                        })

        self.samples = samples

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx):
        sample = self.samples[idx]
        image_path = sample["path"]
        label = sample["label"]

        image = Image.open(image_path).convert("RGB")
        image = image.resize((self.image_size, self.image_size))
        image = np.array(image)

        mask = None  # 此任务中 mask 可为空

        if self.common_transform:
            out = self.common_transform(image=image)
            image = out['image']
        if self.post_transform:
            out = self.post_transform(image=image)
            image = out['image']

        output = {
            "image": image,
            "label": torch.tensor(label).long()
        }
        # Apply post-processing functions
        if self.post_funcs:
            self.post_funcs(output)

        return output

    def __str__(self):
        label_counts = {0: 0, 1: 0}
        for sample in self.samples:
            label = sample["label"]
            if label in label_counts:
                label_counts[label] += 1
            else:
                label_counts[label] = 1  # 处理非0/1情况

        return (f"DeepfakeCrossDataset from config: {self.config_file}\n"
                f"Root path: {self.path}\n"
                f"Compression: {self.compression}\n"
                f"Split mode: {self.split_mode}\n"
                f"Total samples: {len(self.samples)}\n"
                f"Label 0 (real) samples: {label_counts.get(0, 0)}\n"
                f"Label 1 (fake) samples: {label_counts.get(1, 0)}")


if __name__ == '__main__':
    dataset = DeepfakeCrossDataset(path='/mnt/data1/public_datasets/Deepfake',
                                   config_file='/mnt/data1/public_datasets/Deepfake/UADFV.json', split_mode='test')

    dataset.__getitem__(0)

    import pdb;

    pdb.set_trace()
    # path不变  唯一要变的是config_file  c40那个数据集compression变成'c40' 前面是名字 以及split_mode变为train或者test
    # 训练:FaceForensics++ /mnt/data1/public_datasets/Deepfake/FaceForensics++.json compression='c23'
    # 测试: 1. FaceForensics++  /mnt/data1/public_datasets/Deepfake/FaceForensics++.json
    # dataset = DeepfakeCrossDataset(path='/mnt/data1/public_datasets/Deepfake', config_file='/mnt/data1/public_datasets/Deepfake/FaceForensics++.json', compression='c40')
    # 2. FaceForensics++_40  /mnt/data1/public_datasets/Deepfake/FaceForensics++.json compression='c40'
    # dataset = DeepfakeCrossDataset(path='/mnt/data1/public_datasets/Deepfake', config_file='/mnt/data1/public_datasets/Deepfake/FF-DF.json',split_mode='test') 
    # len(dataset) == 8952
    # 3. FF-DF /mnt/data1/public_datasets/Deepfake/FF-DF.json  test
    # 4. FF-F2F /mnt/data1/public_datasets/Deepfake/FF-F2F.json test
    # 5.FF-FS /mnt/data1/public_datasets/Deepfake/FF-FS.json test
    # 6. FF-NT  /mnt/data1/public_datasets/Deepfake/FF-NT.json test
    # 7. Celeb-DF-v1 /mnt/data1/public_datasets/Deepfake/Celeb-DF-v1.json test
    # 8.Celeb-DF-v2 /mnt/data1/public_datasets/Deepfake/Celeb-DF-v2.json test
    # 9. DFD /mnt/data1/public_datasets/Deepfake/DeepFakeDetection.json test
    # 10. DFDC /mnt/data1/public_datasets/Deepfake/DFDC.json test
    # 11. DFDCP /mnt/data1/public_datasets/Deepfake/DFDCP.json test
    # 12. FaceShifter /mnt/data1/public_datasets/Deepfake/FaceShifter.json test
    # 13. UADFV /mnt/data1/public_datasets/Deepfake/UADFV.json test
